#pragma inline
#include <dos.h>
#include <alloc.h>
#include <stdio.h>
#include <conio.h>
#include <fcntl.h>
#include <io.h>
#include <wgt.h>

#define NORMAL 0
#define PARALLAX 1
#define MAXWINDOWS 4

/* Please note that the default tile when you first run the map maker is
   tile 201. This is remarked out in this version since we no longer write
   to 320x200 blocks. In order to do so, we would need to rewrite the bar 
   command to write with different widths.  I'm saving that for a future
   release in which I'll change all of the procedures, so you can draw on
   blocks of any width! The blocks are now only as big as you need them,
   to decrease memory usage.  If you have the default tile in any or your maps, you should
   scan through the map and replace any occurences of this tile with a new one,
   right after you load the map in.
*/

/**************************************************************************
 WordUp Graphics Toolkit source Copyright 1993 WordUp Software Productions
 Licensed for personal use only

 Written by:
			 Chris Egerter
			 94 Andover Drive
			 London Ontario, Canada
			 N6J 3X2
			 (519)472-3661

Belongs to library: WGT4SCR.LIB

***************************************************************************/

/*  		      Multidirectional Scrolling Library 

Naturally this is the most complicated set of routines in the toolkit.
I don't think you'll want to touch much of this.  You probably won't be
able to change anything unless you understand all of the code perfectly,
and even I don't know it. It took me over two months to develop, and
there are still bugs in it.

Here is the basic idea:

Let's begin with a pattern of tiles on the screen.

****************	Suppose we want to scroll the screen opposite
*T***T****H***HH 	to the left. Some scrollers simply shift the
*T***T****H*****	screen in the direction, and draw the new tiles
*TTTTT**HHHHH***	at the edges as needed. You would then need to
********H***H***	copy the result to another virtual page, put on
********H***H***	the sprites, and then finally copy it to the visual
   Frame   1		screen. This means there are 3 copy operations.

***************-	To the left is the screen shifted over by one tile.
T***T****H***HH-	WGT uses a faster method of copying, which only uses
T***T****H*****-	two copy operations.  Raw scroll data is stored on
TTTTT**HHHHH***-	block scroll2.  This screen is copied in 4 different
*******H***H***-	chunks over to different positions on block scroll1.
*******H***H***-	The sprites are then pasted on, and the whole window
    Frame 2    ^	is copied to the visual screen.
	       |
	       The dashes represent the new area which is scrolled on.

While this is a much more complicated method than the first, it requires
one less full screen copy.  When an area of the screen scrolls, notice how
most of the area does not change.  Only the area at the edges need to
be redrawn.  This is why redrawing all of the tiles on the screen does not
make sense.  Why redraw all of the tiles when only two of the edges change
at the most? (you can't scroll left and right or up and down at the same
time, only up left or down right, etc)

By keeping a coordinate placement and increasing the coordinate, we can
easily make an area scroll across the screen with two copy operations
The ^ shows where the placement is.  In each frame the area starting
at ^ to the end is copied to the first column in the result frame.
The columns before the ^ are copied to the column at the ^.
In other words, the screen is split into two sections at the placement,
and both are copied to the opposite sides of the screen.

Raw frames: (scroll2)
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
^            ^            ^            ^            ^
| 	     |    	---   	    ----  	-----
|	    |           |           |           |    
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
Result frames (scroll1)

As you can see, the screen appears to scroll left as the placement
moves right. The only problem is the scroll wraps around the side,
and infinitely scrolls left. To make new tiles scroll appear on the
right side, we just draw new ones onto the raw frame, at one column
before the placement. Hopefully you can follow this description, 
because it gets much more complicated than that...

Of course a game that scrolls by tiles at a time wouldn't be very
smooth. To scroll by one pixel at a time, you need to increase the
placement by one, and only draw on 1 column of the new tiles.
WGT does this by keeping a block with two tiles across (one column 
for the left side, and one for the right).  Every sixteen pixels that
you scroll, the sides are updated, and new tiles are drawn on these
temporary blocks.  This means only the new tiles on the side are drawn
one out of every 16 pixels scrolled.  This is obviously much faster
than drawing the tiles on every time, since most of the time needed
to draw a single frame is just using memcpy about 5-6 times. 

This is the basic idea of multidirectional scrolling the WGT way.
The above example only accounts for one direction. WGT also takes
care of the up and down scrolling as well, in the similar manner.
The only difference between the above example and what is used here, is
the fact that the screen is copied in four sections, not two.
(Because we're combining up/down and left/right)

Sprites do not need to have their background saved, since in each frame
we are copying the whole scroll window over from scroll2. The sprites
can be drawn on scroll1, and will be erased the next time the window
is updated.


WARNING! This concludes the beginning explanation.  It gets much harder
to understand after this point! I am not responsible if you fry your
brain trying to figure out what everything does here...
You can modify it if you want, but the code is very touchy, and you
will probably get graphics messed up all over the screen if you change
something critical. (Believe me, I've tried it!)

*/

/* users should know about these */
block *scrolltiles[4];
block scroll2[MAXWINDOWS];  /* raw scroll data - looks messed up if viewed normally */
block scroll1[MAXWINDOWS];  /* assembled scroll data, copied from scroll2, but in different
		      locations. The sprites are drawn overtop onto this block */
block udside[MAXWINDOWS];   /* A block with dimensions windowmaxx+1 by 32
		      The extra 1 in windowmaxx+1 is because my math isn't right
		      somewhere.... 

		      Contains the top and bottom rows of tiles which appear
		      next according to the worldy position.
		      Tile are organized like this :  123456789...  (top)
						      123456789...  (bottom) */

block lrside[MAXWINDOWS];   /* A block with dimensions 32 by windowmaxy+1
		      Contains the left and right columns of tiles which appear
		      next according to the worldx position.
		      Tile are organized like this :  11
						      22
						      33
						      44
						      .. 
					   left-------^^------right  */

int mapwidth[MAXWINDOWS],mapheight[MAXWINDOWS]; /* the width and height (in tiles) of the current map loaded */
int tilewidth[MAXWINDOWS],tileheight[MAXWINDOWS]; /* the width and height (in pixels) of tiles */


int spritewidth[1001],spriteheight[1001]; /* width and height of sprites
					     so we don't need to recalc all the time */
int numsprites;				  /* # of sprites to show with wshowobjects  */

typedef struct {		 /* object structure */
	char on;		 /* sprite is turned on=1 */
	int x;			 /* world x coordinate (max 320*16=5120) */
	int y;			 /* world y coordinate (max 200*16=3200) */
	unsigned int num;	 /* sprite # from sprites[1001] to show  */
	} scrollsprites;
scrollsprites wobject[1001];	 /* make a 1000 (1..1000) objects */

char tiletype[201];		 /* tile attributes */
int wgtparallax;
int numwindows=1;

int windowminx[MAXWINDOWS],windowminy[MAXWINDOWS],windowmaxx[MAXWINDOWS],windowmaxy[MAXWINDOWS]; /* coords of scrolling area */
				 /* min coords are always 0 for now */

int screenwidth[MAXWINDOWS];  		 /* width of scrolling window */
int worldx[MAXWINDOWS]={0,0},worldy[MAXWINDOWS]={0,0}; /* world coordinates of top left corner of
				       scrolling window */

int worldmaxx[MAXWINDOWS]={320,320},worldmaxy[MAXWINDOWS]={320,320}; /* world max x and y for window
				 These are calculated so you can't scroll
				 past the end of the world, depends on
				 window size */
int windowwidth[MAXWINDOWS],windowheight[MAXWINDOWS];    /* Size of window (in tiles) */

/* A copyright statement to store in executables  */
char *copyrscroll="WordUp Graphics Toolkit:Multidirectional Scrolling Library Copyright 1993 WordUp Software Productions";


/* externals */
extern block sprites[1001];			/* sprite array */
//extern block blocks[1001];			/* tile array */
extern int kbdon[129];				/* keyboard array (on/off) */
/* extern int kbdscanlist[129];	*/		/* used for old method */
extern int tx,ty,bx,by;				/* clipping vars */
				/* top x, top y, bottom x, bottom y */

void wputanywhere (int, int, block, block, int); /* puts a block onto another 
		block of ANY size, not limited to full screen 320x200 blocks
		used for showing sprites onto scroll1 */

void wputtileanywhere (int, int, block, block); /* puts a block onto another
		block of ANY size, not limited to full screen 320x200 blocks
		used for showing sprites onto scroll1 */

/* private stuff */
static int wgtscrollv3,wgtscrollv4; /* misc math for scrolling */
static int wgtscrollv5;		    /* misc var for math routines */
static int x_repeat,y_repeat; 	    /* these are used to store the amount
	 to scroll if the side tile blocks need to be updated halfway
	 through the update. If the side tiles are already halfway copied,
	 and we're scrolling by 16 pixels, 8 pixels of the current tiles
	 must be copied, and then the new tiles must be redrawn and 8
	 more pixels copied. */

static int orig_x_speed,orig_y_speed; /* stores original speeds in case
			 the condition above with x_repeat occurs */
static int redrawside[MAXWINDOWS];  /* if wputworldblock is used, all of the side tiles
		 are redrawn, in case they need to be updated. This is
		 probably why animating tiles slow the scrolling down a lot... */

static int origkey;	/* original location in keyboard buffer */
static int intkey;	/* current location in keyboard buffer */
static int cabsx,cabsy; /* absolute value of the speedx and speedy vars */

static int column_tile_loc[MAXWINDOWS]={0,0}; /* keeps track of the position within
			   lrside to copy the side tiles from.
			   Each position increases or decreases with the
			   scroll speed, and wrap around at 0 or 16 */

static int row_tile_loc[MAXWINDOWS]={0,0}; /* same as column_tile_loc only for upside. */

static int scrlly[MAXWINDOWS]={0,0},scrllx[MAXWINDOWS]={0,0};   /* offset for splitting screen
				 see coordinate placement in example above */

static int upredraw[MAXWINDOWS],downredraw[MAXWINDOWS],leftredraw[MAXWINDOWS],rightredraw[MAXWINDOWS];
		/* tells whether the tiles need to be redrawn on on of the
		 sides */
static int speedx,speedy; /* stores the scrolling speeds */

static void wqcopyscreen (int, int, int, int, block ptr, int, int, block);
		/* a quick copyscreen */
static void wcopyside (int, int, int, int, block ptr, int, int, block);
		/* copies the side tiles onto the main scrolling data window */

static void scl (int);  /* copies the appropriate side depending on */
static void scr (int);  /* the scroll speeds */
static void scu (int);
static void scd (int);

void wputtile (int, int, block, block); /* puts a tile onto a screen */

static void interrupt kbd(); /* interrupt prototype */
static void interrupt (*oldkeyboard)(); /* interrupt function pointer */
void wcopyanywhere (int, int, int, int, block, int, int, block);


/* funcs */
void installkbd (void); 		/* start up keyboard handler */
void winitscroll (int, int,int,int,int,int,block *);	/* defines window size, and sets up vars */
void wshowwindow (int, int, wgtmap,int); 	/* draw the tiles on at a beginning place
					   and sets up more vars */
void wscrollwindow (int, int, wgtmap,int);  /* scroll the screen
void wshowobjects (void);	           draw the objects onto scroll1 */
int  wgetworldblock (int, int, wgtmap); /* returns the tile number at a
					   world coordinate */
void wputworldblock (int, int, int, wgtmap,int); /* puts a tile at a world coordinate */
void wcopyscroll (int, int,int);	/* copies scroll1 onto visual screen */
void wendscroll (int);		/* frees blocks, and shuts down scrolling */
void uninstallkbd (void);  	/* stop keyboard handler */
void wfreemap (wgtmap);		/* deallocates memory for a map */
int soverlap (int, int);	/* sees if two objects overlap */
int currentwindow=0;		/* current active scrolling window
				   used for multi window games */
int windowmode[MAXWINDOWS];     /* Mode for each window (norm or para) */


wgtmap wloadmap (char *mapfile, int currentwindow)
/* Loads a map file into memory */
{
    wgtmap ptr,orig;
    int ctr,numread,mapver; 
    unsigned char conv;
    unsigned size,bytes;
    
    if (wgtlibrary == NULL)
    {
        if ((libf = fopen (mapfile, "rb")) == NULL)
            return NULL;
    }
    else 
    {
        if ((libf = fopen (wgtlibrary, "rb")) == NULL)
            return NULL;
        readheader ();
        findfile (mapfile);
        if (lresult == 1)
            fseek (libf, lfpos, SEEK_SET);
        if (checkpassword (password) == 0)
        {
            textmode (C80);
            printf ("Incorrect password");
            exit (1);
        }
    }
    
    if ((wgtlibrary != NULL) & (lresult == 0)) goto mapstop;
    
    mapver = getw (libf); /* read in version number */
    if ((mapver != 8972) & (mapver != 8973))
    /* not a map file */
    {
        textmode (C80);
	printf ("Invalid map header: %i\n", mapwidth[currentwindow]);
        exit (1);
    }
    if (mapver == 8972)
    /* first version */
    {
	mapwidth[currentwindow] = fgetc (libf);
	mapheight[currentwindow] = fgetc (libf);
    /* only allowed for 200x200 maps */
    }
    else
    /* second version */
    {
	mapwidth[currentwindow] = getw (libf);
	mapheight[currentwindow] = getw (libf);
    /* allows for 320x200 maps, therefore integer needed */
    }
    
    size = mapheight[currentwindow]*mapwidth[currentwindow];
    /* calculate size in bytes of map */
    if (size < 65533) 
    {
	ptr = farmalloc (size);			/* allocate memory */

	if (ptr == NULL) return NULL;		/* out of memory? */
	orig = ptr;				/* original ptr stored */
	if (fread (ptr, size, 1, libf) < 1) 	/* read in the data */
	{
	    printf ("Read Failed (MAP).\n");
	    exit (1);                             
	}
	for (ctr = 1; ctr <= 200; ctr++)	/* read in 200 bytes of tile types */
	    tiletype[ctr] = getc (libf);
	ctr = getw (libf);			/* number of objects stored in file  */
	if ((mapver == 8973) & (ctr>0))		        /* second version */
	    fread (&wobject[1], 7, ctr, libf);  /* read ctr object structures */
	else  /* first version */
	{
	    for (numread = 1; numread <= ctr; numread++)
	    {
		fread (&wobject[numread].on, 1, 1, libf); /* on/off */
		fread (&wobject[numread].x, 2, 1, libf);  /* coords */
		fread (&wobject[numread].y, 2, 1, libf);
		fread (&conv, 1, 1, libf);                /* sprite num */
		wobject[numread].num = (unsigned int)conv;
            }
        }
    }
    mapstop:
    ;
    fclose (libf);
    return orig;
}


void wsavemap (char *fn, wgtmap map)
/* Saves a map file using current object positions
   Useful for saving games in the middle of playing a map
   so you don't have to save games at the start of a level only */
{
    FILE *fp;
    wgtmap ptr,orig;
    int ctr,numread,mapver,ynum;
    unsigned char conv;
    unsigned size,bytes;
    
    fp = fopen (fn, "wb");  /* open the file */
    putw (8973, fp);	    /* write the magic number */
    putw (mapwidth[currentwindow], fp);	/* write dimensions */
    putw (mapheight[currentwindow], fp);
    
    size = mapheight[currentwindow]*mapwidth[currentwindow]; /* find size */
    fwrite (map, size, 1, fp); /* write size bytes of data from map ptr */
    for (ctr = 1; ctr <= 200; ctr++)
	fputc (tiletype[ctr], fp);     /* write 200 bytes of tile types */
    
    ynum = 0;
    for (numread = 1; numread < 1001; numread++)   
	if (wobject[numread].num > 0) ynum = numread; /* find highest object number used */
    putw (ynum, fp);
    if (ynum > 0) fwrite (&wobject[1], 7*ynum, 1, fp); /* write ynum object structures */
    fclose (fp);
}


int soverlap (int s1, int s2)
/* Sees if two objects overlap, by comparing the rectangles each are
   contained in.  Pixel precise detection is NOT possible with WGT
   routines. */
{
    int n1,n2;
    
    if ((wobject[s2].on & wobject[s1].on == 1)) /* Are they both on? */
    {
	n1 = wobject[s1].num;  /* for easier reading */
        n2 = wobject[s2].num;
        if (( wobject[s2].x >= wobject[s1].x - spritewidth[n2] ) &&
		( wobject[s2].x <= wobject[s1].x + spritewidth[n1] ) &&  /* check all four corners */
		( wobject[s2].y >= wobject[s1].y - spriteheight[n2] ) &&
		( wobject[s2].y <= wobject[s1].y + spriteheight[n1] )) return 1;
    }
    return 0; /* not colliding */
}

void wscrollwindow (int wsx, int wsy, wgtmap scrollmap,int currentwindow)
/* Scroll a window using scrollmap by wsx pixels left/right
   and wsy up/down. */
{
    wgtmap tempmap;
    int q;		/* counter */
    block *blocks;

    blocks=scrolltiles[currentwindow];
    wgtparallax=windowmode[currentwindow];

    speedx = wsx;       /* store speeds into global vars for support routines */
    speedy = wsy;

    abuf = MK_FP (FP_SEG (scroll2[currentwindow] + 4), FP_OFF (scroll2[currentwindow] + 4));
    /* aka wsetscreen(scroll2); */

    if (speedx > tilewidth[currentwindow]) speedx = tilewidth[currentwindow];	/* can't scroll more than 16 pixels (1 tile) */
    else if (speedx < - tilewidth[currentwindow]) speedx = - tilewidth[currentwindow];	/* at a time */
    if (speedy > tileheight[currentwindow]) speedy = tileheight[currentwindow];
    else if (speedy < - tileheight[currentwindow]) speedy = - tileheight[currentwindow];
    
    if ((x_repeat) | (y_repeat))  /* if this wscrollwindow is recursed */
    {
	speedx = x_repeat;  /* set speed to remainder of first call */
	speedy = y_repeat;
    }

    x_repeat = 0; orig_x_speed = speedx; /* store original speeds */
    y_repeat = 0; orig_y_speed = speedy;
    /* clear out previous remainder so wscrollwindow won't keep calling itself */
	
    
    if (speedx)		/* changing the x coordinate */
    {
	upredraw[currentwindow] = 1; downredraw[currentwindow] = 1;
	/* Redraw top and bottom tiles, regardless of tile position */
		
 
	/* if you hit the edge of the world, decrease the scrolling speed */
	if (worldx[currentwindow] + speedx < 0)
	    speedx += abs (worldx[currentwindow] + speedx);
	else if (worldx[currentwindow] + speedx > worldmaxx[currentwindow])
	    speedx -= abs (worldmaxx[currentwindow] - (worldx[currentwindow] + speedx));
	

	if ((column_tile_loc[currentwindow] + speedx > tilewidth[currentwindow]) && (column_tile_loc[currentwindow] != tilewidth[currentwindow]))
	/* if side tile location must wrap around */
	{
	    /* set speed to end move at tile edge */
	    speedx = tilewidth[currentwindow] - column_tile_loc[currentwindow];

	    /* set tile location to right edge */
	    column_tile_loc[currentwindow] = tilewidth[currentwindow];
	    x_repeat = orig_x_speed - speedx; /* get speed for recursive call */
	}
	else if ((column_tile_loc[currentwindow] + speedx < 0) && (column_tile_loc[currentwindow] != 0))
	/* wrap around at 0 */
	{
	    speedx = - column_tile_loc[currentwindow];
	    /* set tile location to left edge */
	    column_tile_loc[currentwindow] = 0;
	    x_repeat = orig_x_speed - speedx;
        }
        else  
        {
	    /* tile position doesn't wrap around */
	    column_tile_loc[currentwindow] += speedx;

	    /* tile needs to be redrawn */
	    if (column_tile_loc[currentwindow] < 0)
            {
		column_tile_loc[currentwindow] += tilewidth[currentwindow]; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
	    }
	    else if (column_tile_loc[currentwindow] > tilewidth[currentwindow])
	    {
		column_tile_loc[currentwindow] -= tilewidth[currentwindow]; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
            }
        }
        
        
	worldx[currentwindow] += speedx; /* Update world coordinates */
	scrllx[currentwindow] += speedx; /* Update x scroll position */


	/* wrap around scrllx */
	if ((speedx < 0) && (scrllx[currentwindow] < windowminx[currentwindow]))
	    scrllx[currentwindow] += windowmaxx[currentwindow] + 1;
	else if ((speedx > 0) && (scrllx[currentwindow] > windowmaxx[currentwindow]))
	    scrllx[currentwindow] -= windowmaxx[currentwindow] + 1;
        
    }
    
    if (speedy) /* moving up or down */
    {
	leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
	/* Redraw left and right tiles, regardless of tile position */
		

	/* if you hit the edge of the world, decrease the scrolling speed */
	if (worldy[currentwindow] + speedy < 0)
	    speedy += abs (worldy[currentwindow] + speedy);
	else if (worldy[currentwindow] + speedy > worldmaxy[currentwindow])
	    speedy -= abs (worldmaxy[currentwindow] - (worldy[currentwindow] + speedy));
        
	if ((row_tile_loc[currentwindow] + speedy > tileheight[currentwindow]) && (row_tile_loc[currentwindow] != tileheight[currentwindow]))
	/* if side tile location must wrap around
	   wrap around at 16 */
	{
	    /* set speed to end move at tile edge */
	    speedy = tileheight[currentwindow] - row_tile_loc[currentwindow];

	    /* set tile location to bottom edge */
	    row_tile_loc[currentwindow] = tileheight[currentwindow];
	    y_repeat = orig_y_speed - speedy; /* get speed for recursive call */
	}
	else if ((row_tile_loc[currentwindow] + speedy < 0) && (row_tile_loc[currentwindow] != 0))
	/* wrap around at 0 */
	{
	    /* set speed to end move at tile edge */
	    speedy = - row_tile_loc[currentwindow];

	    /* set tile location to top edge */
	    row_tile_loc[currentwindow] = 0;
	    y_repeat = orig_y_speed - speedy; /* get speed for recursive call */
        }
        else
	  /* tile position doesn't wrap around */
	{
	    row_tile_loc[currentwindow] += speedy;

	    if (row_tile_loc[currentwindow] < 0)
	    /* tile needs to be redrawn */
	    {
		row_tile_loc[currentwindow] += tileheight[currentwindow]; upredraw [currentwindow]= 1; downredraw[currentwindow] = 1;
	    }
	    else if (row_tile_loc[currentwindow] > tileheight[currentwindow])
	    {
		row_tile_loc[currentwindow] -= tileheight[currentwindow]; upredraw[currentwindow] = 1; downredraw[currentwindow] = 1;
            }
        }
        
        
	worldy[currentwindow] += speedy;	/* Update world coordinates */
	scrlly[currentwindow] += speedy;	/* Update scroll y position */

	/* wrap around scrlly */
	if ((speedy < 0) && (scrlly[currentwindow] < windowminy[currentwindow]))
	    scrlly[currentwindow] += windowmaxy[currentwindow] + 1;
	else if ((speedy > 0) && (scrlly[currentwindow] > windowmaxy[currentwindow]))
	    scrlly[currentwindow] -= windowmaxy[currentwindow] + 1;
        
    }
    
    
    
    if (((speedx > 0) && (rightredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    /* See if we need to redraw the right hand tiles */
    {
	if (redrawside[currentwindow] == 0) rightredraw[currentwindow] = 0;

	tx = tilewidth[currentwindow]; ty = 0; bx = tilewidth[currentwindow]*2+1; by = windowmaxy[currentwindow];
	// used to be tx 16 bx 35

	/* Set clipping for side tile */

	wgtscrollv4 = worldy[currentwindow] % tileheight[currentwindow];
	if (wgtscrollv4 != 0)  /* Not even with a tile */
	    wgtscrollv4 = 0 - wgtscrollv4; /* Make the tile partly off the top of the screen */
	/* Find the first y coordinate to put tiles */

	tempmap = scrollmap;  /* Beginning of map data */
	tempmap += ( ((worldy[currentwindow]) / tileheight[currentwindow])*mapwidth[currentwindow] + (((worldx[currentwindow] - 1) / tilewidth[currentwindow]) + windowwidth[currentwindow]));
	/* Jump to offset of tiles needed */

	/* Now draw a column of tiles on the right side */
	for (q = 0; q <= windowheight[currentwindow] + 1; q++)
	{
	    wgtscrollv3 = *tempmap; /* Get the tile */

	    /*  if (wgtscrollv3==201)
		 { wsetcolor(0); wbar(19,wgtscrollv4,34,wgtscrollv4+15); }
	     Used for tile 201, taken out for version 4.0 */

	    wputanywhere (tilewidth[currentwindow], wgtscrollv4, blocks[wgtscrollv3], lrside[currentwindow],0);
	    /* Put the block onto the right hand side block */

	    wgtscrollv4 += tileheight[currentwindow]; /* Add the height of one tile */
	    tempmap += mapwidth[currentwindow]; /* Move down one row */
	}
	tx = 0; ty = 0; bx = 319; by = 199;
	/* Reset clipping */
    }
    
    if (((speedx < 0) && (leftredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    /* See if we need to redraw left hand tiles */
    {
	if (redrawside [currentwindow]== 0)leftredraw[currentwindow] = 0;

	tx = 0; ty = 0; bx = tilewidth[currentwindow]; by = windowmaxy[currentwindow];
	/* Set clipping for side tile */

	wgtscrollv4 = worldy[currentwindow] % tileheight[currentwindow];
	if (wgtscrollv4 != 0)   /* Not even with a tile */
	    wgtscrollv4 = 0 - wgtscrollv4; /* Make it partly off the top */
	/* Find the first y coordinate to put tiles */

	tempmap = scrollmap; /* Start of map data */
	tempmap += ( ((worldy[currentwindow]) / tileheight[currentwindow])*mapwidth[currentwindow] + (worldx[currentwindow] / tilewidth[currentwindow]));
	/* Jump to offset of the tile we want */

	/* Now draw a column of tiles on the left side */
	for (q = 0; q <= windowheight[currentwindow] + 1; q++)
	{
	    wgtscrollv3 = *tempmap;	/* Get tile */
	    /*  if (wgtscrollv3==201)
		 { wsetcolor(0); wbar(3,wgtscrollv4,18,wgtscrollv4+15); }
	     Used for background tile 201, removed in version 4.0 */

	    wputanywhere (0, wgtscrollv4, blocks[wgtscrollv3], lrside[currentwindow],0);
	    /* Put a tile onto the side tile block */

	    wgtscrollv4 += tileheight[currentwindow]; /* Move down one tile */
	    tempmap += mapwidth[currentwindow]; /* Jump down one row */
	}
	tx = 0; ty = 0; bx = 319; by = 199;
	/* Reset clipping */
    }
    
    if (((speedy > 0) && (downredraw[currentwindow] != 0)) | (redrawside[currentwindow] == 1))
    /* If we need to redraw the bottom tiles */
    {
	if (redrawside[currentwindow] == 0) downredraw[currentwindow] = 0;
	tx = 0; ty = tileheight[currentwindow]; bx = windowmaxx[currentwindow]; by = tileheight[currentwindow]*2+1;
	/* Clip to bottom tile block */

	wgtscrollv4 = worldx[currentwindow] % tilewidth[currentwindow];
	if (wgtscrollv4 != 0)  /* Not even with a tile */
	    wgtscrollv4 = 0 - wgtscrollv4; /* Make tile partly off side of screen */

	tempmap = scrollmap; /* Reset to start of map */
	tempmap += ( (((worldy[currentwindow] - 1) / tileheight[currentwindow]) + windowheight[currentwindow])*mapwidth[currentwindow] + ((worldx[currentwindow]) / tilewidth[currentwindow]));
	/* Jump to the first tile we want */

	/* Now draw a row of tiles on the bottom */
	for (q = 0; q <= windowwidth[currentwindow] + 1; q++)
	{
	    wgtscrollv3 = *tempmap;   /* Get tile */
	    
	    /*  if (wgtscrollv3==201)
		 { wsetcolor(0); wbar(wgtscrollv4,windowmaxy+19,wgtscrollv4+15,windowmaxy+34); }
	     Used for background tile 201, removed in version 4.0 */


	    /* Put a tile onto the bottom tile block */
	    wputanywhere (wgtscrollv4, tileheight[currentwindow], blocks[wgtscrollv3], udside[currentwindow],0);

	    wgtscrollv4 += tilewidth[currentwindow]; /* Move right one tile */
	    tempmap++;         /* Go to next tile */
	}
	tx = 0; ty = 0; bx = 319; by = 199;
	/* Reset clipping */
    }
    
    if (((speedy < 0) && (upredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    /* If we need to redraw the top tiles */
    {
	if (redrawside[currentwindow] == 0)upredraw[currentwindow] = 0;
	tx = 0; ty = 0; bx = windowmaxx[currentwindow]; by = tileheight[currentwindow];
	/* Clip to top tile block */

	wgtscrollv4 = worldx[currentwindow] % tilewidth[currentwindow];
	if (wgtscrollv4 != 0)  /* Not even with a tile */
	    wgtscrollv4 = 0 - wgtscrollv4; /* Make it partly off the side */

	tempmap = scrollmap; /* Reset to the start of map data */
	tempmap += ( ((worldy[currentwindow] / tileheight[currentwindow])*mapwidth[currentwindow]) + ((worldx[currentwindow]) / tilewidth[currentwindow]));
	/* Jump to the first tile we want */

	/* Now draw a row of tiles */
	for (q = 0; q <= windowwidth[currentwindow] + 1; q++)
	{
	    wgtscrollv3 = *tempmap; /* Get the tile */
	    /*  if (wgtscrollv3==201)
		 { wsetcolor(0); wbar(wgtscrollv4,windowmaxy+3,wgtscrollv4+15,windowmaxy+18); }
	     Removed for version 4.0 */

	    /* Put a tile onto the side tile block */
	    wputanywhere (wgtscrollv4, 0, blocks[wgtscrollv3], udside[currentwindow],0);
	    wgtscrollv4 += tilewidth[currentwindow];  /* Move right one tile */
	    tempmap++;  /* Get next tile */
	}
	tx = 0; ty = 0; bx = 319; by = 199;
	/* Reset clipping */
    }
    
    wsetscreen (scroll2[currentwindow]);
    /* Go to raw data screen */

    cabsx = abs (speedx);
    cabsy = abs (speedy);
    /* Used many times below so only calculate once */
	
    /* Now copy the side tiles onto the raw data screen */
    if (speedx > 0)     /* Copy right side */
	scr(currentwindow);
    else if (speedx < 0)/* Copy left side */
	scl(currentwindow);
    if (speedy > 0)     /* Copy bottom side */
	scd(currentwindow);
    else if (speedy < 0)/* Copy top side */
	scu(currentwindow);
   
    tx = 0; ty = 0; bx = 319; by = 199;
    /* Reset clipping */

    if ((x_repeat == 0) && (y_repeat == 0))
    /* Don't copy the whole screen if we need to recurse, since will
       we be copying it the second (final) time. */
    {
	wgtscrollv3 = windowmaxx[currentwindow] - scrllx[currentwindow] + 1;
	wgtscrollv4 = scrllx[currentwindow] - 1;
	/* Common variables for below */
		
	/* This is the main scrolling data manipulation code.
	   It takes the data from scroll2 (raw data) and assembles it
	   into the scrolling background you see when playing.
	   It assembles the data by copying from scroll2 onto scroll1
	   in a different position. */


	if (wgtparallax==0)
	{
	wqcopyscreen (windowminx[currentwindow], windowminy[currentwindow], wgtscrollv4, scrlly[currentwindow], scroll2[currentwindow], wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	/* 1st quadrant */
	wqcopyscreen (scrllx[currentwindow], windowminy[currentwindow], windowmaxx[currentwindow], scrlly[currentwindow] - 1, scroll2[currentwindow], windowminx[currentwindow], windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	/* 2nd quadrant */
	wqcopyscreen (windowminx[currentwindow], scrlly[currentwindow], wgtscrollv4, windowmaxy[currentwindow], scroll2[currentwindow], wgtscrollv3, windowminy[currentwindow], scroll1[currentwindow]);
	/* 3rd quadrant */
	wqcopyscreen (scrllx[currentwindow], scrlly[currentwindow], windowmaxx[currentwindow], windowmaxy[currentwindow], scroll2[currentwindow], windowminx[currentwindow], windowminy[currentwindow], scroll1[currentwindow]);
	/* 4th quadrant */
	}
	else
	 {
	wcopyanywhere (windowminx[currentwindow], windowminy[currentwindow], wgtscrollv4, scrlly[currentwindow], scroll2[currentwindow], wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	/* 1st quadrant */
	wcopyanywhere (scrllx[currentwindow], windowminy[currentwindow], windowmaxx[currentwindow], scrlly[currentwindow] - 1, scroll2[currentwindow], windowminx[currentwindow], windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	/* 2nd quadrant */
	wcopyanywhere (windowminx[currentwindow], scrlly[currentwindow], wgtscrollv4, windowmaxy[currentwindow], scroll2[currentwindow], wgtscrollv3, windowminy[currentwindow], scroll1[currentwindow]);
	/* 3rd quadrant */
	wcopyanywhere (scrllx[currentwindow], scrlly[currentwindow], windowmaxx[currentwindow], windowmaxy[currentwindow], scroll2[currentwindow], windowminx[currentwindow], windowminy[currentwindow], scroll1[currentwindow]);
	/* 4th quadrant */
	}

    }
    
    if ((x_repeat) | (y_repeat))
    /* If we need to recurse, call it again with remaining distances */
	wscrollwindow (x_repeat, y_repeat, scrollmap,currentwindow);

    redrawside[currentwindow] = 0;
    /* Reset drawing sides */
}

void scl (int currentwindow)
/* Copies the left side tile onto the raw data screen (scroll2) */
{
    wgtscrollv3 = column_tile_loc[currentwindow];
    wgtscrollv4 = wgtscrollv3 + cabsx - 1;
    wgtscrollv5 = scrllx[currentwindow];
    /* Common variables for below */


    /* Copy the side tile onto scroll2 */
    wcopyside (wgtscrollv3, 0, wgtscrollv4, windowmaxy[currentwindow] - scrlly[currentwindow], lrside[currentwindow], wgtscrollv5, scrlly[currentwindow], scroll2[currentwindow]);
    if (scrlly[currentwindow] > 0)
	wcopyside (wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, wgtscrollv4, windowmaxy[currentwindow], lrside[currentwindow], wgtscrollv5, 0, scroll2[currentwindow]);
}

void scr (int currentwindow)
/* Copies the right side tile onto the raw data screen (scroll2) */
{
    wgtscrollv3 = tilewidth[currentwindow] + column_tile_loc[currentwindow] - cabsx;
    wgtscrollv4 = wgtscrollv3 + cabsx - 1;
    wgtscrollv5 = scrllx[currentwindow] - cabsx;
    /* Common variables for below */

    /* Check for wraparound */
    if (wgtscrollv5 < 0)
	wgtscrollv5 += windowmaxx[currentwindow] + 1;

    /* Copy the side tile onto scroll2 */
    wcopyside (wgtscrollv3, 0, wgtscrollv4, windowmaxy[currentwindow] - scrlly[currentwindow], lrside[currentwindow], wgtscrollv5, scrlly[currentwindow], scroll2[currentwindow]);
    if (scrlly[currentwindow]> 0)
	wcopyside (wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, wgtscrollv4, windowmaxy[currentwindow], lrside[currentwindow], wgtscrollv5, 0, scroll2[currentwindow]);
}

void scu (int currentwindow)
/* Copies the upper side tile onto the raw data screen (scroll2) */
{
    wgtscrollv3 = row_tile_loc[currentwindow];
    wgtscrollv4 = wgtscrollv3 + cabsy - 1;
    wgtscrollv5 = scrlly[currentwindow];
    /* Common variables for below */
/*    if (wgtscrollv5 > windowmaxy[currentwindow])
	wgtscrollv5 -= windowmaxy[currentwindow] + 1;*/

    /* Copy the side tile onto scroll2 */
    wcopyside (0, wgtscrollv3, windowmaxx[currentwindow] - scrllx[currentwindow], wgtscrollv4, udside[currentwindow], scrllx[currentwindow], wgtscrollv5, scroll2[currentwindow]);
    if (scrllx[currentwindow] > 0)
	wcopyside (windowmaxx[currentwindow] - scrllx[currentwindow] + 1, wgtscrollv3, windowmaxx[currentwindow], wgtscrollv4, udside[currentwindow], 0, wgtscrollv5, scroll2[currentwindow]);
}

void scd (int currentwindow)
/* Copies the bottom side tile onto the raw data screen (scroll2) */
{
    wgtscrollv3 = tileheight[currentwindow] + row_tile_loc[currentwindow] - cabsy;
    wgtscrollv4 = wgtscrollv3 + cabsy - 1;
    wgtscrollv5 = scrlly[currentwindow] - cabsy;
    /* Common variables for below */

    /* Check for wraparound */
    if (wgtscrollv5 < 0)
	wgtscrollv5 += windowmaxy[currentwindow] + 1;

    /* Copy the side tile onto scroll2 */
    wcopyside (0, wgtscrollv3, windowmaxx[currentwindow] - scrllx[currentwindow], wgtscrollv4, udside[currentwindow], scrllx[currentwindow], wgtscrollv5, scroll2[currentwindow]);
    if (scrllx[currentwindow] > 0)
	wcopyside (windowmaxx[currentwindow] - scrllx[currentwindow] + 1, wgtscrollv3, windowmaxx[currentwindow], wgtscrollv4, udside[currentwindow], 0, wgtscrollv5, scroll2[currentwindow]);
}



void interrupt kbd (void)
/* Keyboard press handler
   Handles any number of keypresses at once */
{
    int i,j;
    
    disable ();				/* Disable interrupts */
    i = inportb (0x60);			/* Read a byte from the keyboard port */
    kbdon[i % 128] = (i < 128);  	/* Make 0 or 1 depending on which scan code was returned */
    /* If the code is > 128, the key was released */
    /* If scan codes go higher than 128, then this will store results in more than one array element */
	
    (* oldkeyboard)();
    /* call old interrupt 9 handler            */
    pokeb (0x000, 0x041C, intkey);
    enable ();
}

void installkbd ()
/* Set up the new keyboard handler */
{
    oldkeyboard = getvect (0x09);  /* Get the old keyboard interrupt */
    setvect (0x09, kbd); /* Set the vector to the new interrupt */
    intkey = peekb (0x000, 0x041A); /* Get the position in the keyboard buffer */
    origkey = intkey;
    intkey += 2; /* Go to the next place in buffer */
    if (intkey > 58) /* Wrap around if needed */
        intkey = 30;
}


void uninstallkbd ()
/* Removes our keyboard handler and restores the original one */
{
    setvect (0x09, oldkeyboard);
    pokeb (0x000, 0x041A, origkey); 	    /* Restore the position in the keyboard buffer */
    delay (100); while (kbhit ()) getch (); /* Clear out any keys from the buffer */
}


/***================================================================***
*             Copy a side tile block onto the raw data screen         *
****================================================================***/
void wcopyside (int x, int y, int x2, int y2, block ptr, int desx, int desy, block to)
{
    int width,height,display,maxy;
    int dwidth;
    register int ctr;

    dwidth = wgetblockwidth (ptr);  /* width of block */
    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    display = width; 		    /* Number of pixels to show */
    if ((height > 0) && (display > 0))
    {
	ptr += 4 + x + (y*dwidth); /* Move to offset */
	to += 4;      /* Skip over width and height integers */
	to += desx + (desy*screenwidth[currentwindow]); /* Move to offset */

	for (ctr = 0; ctr < height; ctr++) 
	{
	    memcpy (to, ptr, display); /* Copy the data */
	    ptr += dwidth;	       /* Source increases by the width */
	    to += screenwidth[currentwindow]; /* Dest increases by width of window */
	}
    }
}


/***================================================================***
*            Quick Copy part of a screen to another (no clipping)     *
****================================================================***/
void wqcopyscreen (int x, int y, int x2, int y2, block from, int desx, int desy, block to)
{
    int toadd,fromadd;
    int width,height,display,maxy;
    int damt,wamt,bamt;
    register int ctr;
    int dwidth;                        /* temp window width */
    int swidth;                        /* temp source width */
    block temp;

    if (x2 < 0) x2 = 0;		       /* Check dimensions */
    if (x < 0) x = 0;
    if (y2 < 0) y2 = 0;
    if (y < 0) y = 0;

    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    /* Get width and height */
	
    display = width; /* Number of pixels to copy */
    if ((height > 0) && (display > 0))
    {
	if (from == NULL) /* Visual page */
	{
	    from = MK_FP (0xA000, 0x0000); /* Make a ptr to the VGA screen */
	    swidth = 320;
        }
        else
        {
	    swidth = wgetblockwidth (from); /* Get width of block */
	    from += 4; 			   /* Skip width and height integers */
	}
	if (to == NULL) /* Make a ptr to VGA screen */
	{
	    to = MK_FP (0xA000, 0x0000);
	    dwidth = 320;
	}
	else
        {
	    dwidth = wgetblockwidth (to); /* Set up dest width */
	    to += 4;	/* Skip over width and height integers */
        }
	from += x + (y*swidth);     /* Move to correct offsets */
	to += desx + (desy*dwidth);

	toadd=dwidth-display;
	fromadd=swidth-display;

	damt=display>>2;
	display-=damt<<2;
	wamt=(display>>1);
	display-=(wamt<<1);
	bamt=display;

	if (toadd<0) toadd=0;
	if (fromadd<0) fromadd=0;

	asm push ds;
	asm cld;
	asm lds si,from;
	asm les di,to;
	for (ctr = 0; ctr < height; ctr++)
	{
//	memcpy(to,from,display);
	asm .386
	    asm mov cx,damt;
	    asm rep movsd;
	    asm mov cx,wamt;
	    asm rep movsw;
	    asm mov cx,bamt;
	    asm rep movsb;
	    asm add di,toadd;
	    asm add si,fromadd;
	}
	asm pop ds;
    }
}



/***================================================================***
*            Quick Copy part of a screen to another (no clipping)     *
****================================================================***/
void wshowscroller (int x, int y, int x2, int y2, block from, int desx, int desy, block to)
{
    int width,height,display,maxy;
    register int ctr;
    int dwidth;                        /* temp window width */
    int swidth;                        /* temp source width */
    block temp;
    int toadd,fromadd;


    if (x2 < 0) x2 = 0;		       /* Check dimensions */
    if (x < 0) x = 0;
    if (y2 < 0) y2 = 0;
    if (y < 0) y = 0;

    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    /* Get width and height */
	
    display = width; /* Number of pixels to copy */
    if ((height > 0) & (display > 0))
    {
	if (from == NULL) /* Visual page */
	{
	    from = MK_FP (0xA000, 0x0000); /* Make a ptr to the VGA screen */
	    swidth = 320;
        }
        else
        {
	    swidth = wgetblockwidth (from); /* Get width of block */
	    from += 4; 			   /* Skip width and height integers */
	}
	if (to == NULL) /* Make a ptr to VGA screen */
	{
	    to = MK_FP (0xA000, 0x0000);
	    dwidth = 320;
	}
	else
        {
	    dwidth = wgetblockwidth (to); /* Set up dest width */
	    to += 4;	/* Skip over width and height integers */
        }
	from += x + (y*swidth);     /* Move to correct offsets */
	to += desx + (desy*dwidth);
	toadd=dwidth-display;
	if (toadd<0) toadd=0;
//	fromadd=swidth-display;
	display>>=2;


	asm push ds;
        asm cld;
	asm lds si,from;
	asm les di,to;
	for (ctr = 0; ctr < height; ctr++)
	{
	    asm .386
	    asm mov cx,display;
	    asm rep movsd;
	    asm add di,toadd;
	    asm inc si;
	}
	asm pop ds;
       }
}



void wshowwindow (int numx, int numy, wgtmap scrollmap,int currentwindow)
/* Starts looking at the scrolling world at numx,numy tiles from the
   top left corner  (1,1) is the top corner. */
{
    int i,j;
    wgtmap tempmap;
    block *blocks;

    blocks=scrolltiles[currentwindow];
    wgtparallax=windowmode[currentwindow];
    if ((numx*tilewidth[currentwindow] > worldmaxx[currentwindow]) | (numy*tileheight[currentwindow] > worldmaxy[currentwindow]) | (numx < 0) | (numy < 0))
    /* Not completely within map */
    {
        textmode (C80);
        printf ("\nError: Coordinates outside world in showwindow.");
	printf ("\nCoordinates used: %i,  %i\n", numx, numy);
	printf ("\nMax Coordinates:  %i,  %i\n", worldmaxx[currentwindow], worldmaxy[currentwindow]);
	exit (1);
    }
    
    wsetscreen (scroll2[currentwindow]);
    /* Get ready to set up initial raw data */

    column_tile_loc[currentwindow] = 0; row_tile_loc[currentwindow] = 0;
    /* Reset side tile offsets */

    scrlly[currentwindow] = 0;scrllx[currentwindow] = 0;             /* scrll is ctr for offset */
    /* Reset main positional offsets */

    upredraw[currentwindow] = 1; downredraw[currentwindow] = 1; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
    /* Reset drawing flags */

    wclip (0, 0, 319, 199);

    for (i = 0; i < windowwidth[currentwindow]; i++)
	for (j = 0; j < windowheight[currentwindow]; j++)
    /* Draw a matrix of boxes for initial use */
    {
	tempmap = scrollmap; /* Make a ptr to the map */
	tempmap += (j + numy)*mapwidth[currentwindow] + (i + numx);
	/* Move to the tile we want to display */

	wgtscrollv3 = *tempmap;
	/* Read in the tile */

	if (wgtscrollv3 == 201) /* If it is black  (default empty tile ) */
	    {
	    wsetcolor (0);
	    wbar (windowmaxx[currentwindow] + 19, wgtscrollv4, windowmaxx[currentwindow] + 34, wgtscrollv4 + 15);
            }
	else wputtile (i*tilewidth[currentwindow], j*tileheight[currentwindow], blocks[wgtscrollv3], scroll2[currentwindow]);
    /* Otherwise put the tile onto the raw data screen */
    }
    worldx[currentwindow] = numx*tilewidth[currentwindow];
    worldy[currentwindow] = numy*tileheight[currentwindow];
    /* Reset the initial world coordinates */
}


void winitscroll (int numx, int numy,int mode,int currentwindow,int tw,int th,block *tileref)
/* Initializes scrolling by setting variables, and allocating two screens,
   one called scroll2 for raw scrolling images, and scroll1 for the
   assembled image with sprites on top. */

{
    scrolltiles[currentwindow]=tileref;
    windowmode[currentwindow]=mode;
    tilewidth[currentwindow]=tw;
    tileheight[currentwindow]=th;

    if ((numx > 320/tw) | (numy > 200/th) | (numx < 2) | (numy < 2))
    /* Invalid window size */
    {
        textmode (C80);
        printf ("\nInvalid scroll window in call to initscroll!");
        exit (1);
    }


    windowminx[currentwindow] = 0;
    windowminy[currentwindow] = 0;
    windowmaxx[currentwindow] = numx*tilewidth[currentwindow] - 1;
    windowmaxy[currentwindow] = numy*tileheight[currentwindow] - 1;
    /* coords of scrolling area */

    speedx = 0; speedy = 0;
    screenwidth[currentwindow] = windowmaxx[currentwindow] + 2;

    scroll2[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow] + 1, windowmaxy[currentwindow] + 1);
    if (mode==PARALLAX)
	if (currentwindow==1) scroll1[currentwindow]=scroll1[0];
	else {
	     if (numwindows==1)
	     scroll1[currentwindow]=scroll1[0];
	     else
	     scroll1[currentwindow]=scroll1[2];
	     }
    else {
	scroll1[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow] + 1, windowmaxy[currentwindow] + 1);
	}
    /* get two virtual screens (not necessarily full screen) */

    udside[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow], tileheight[currentwindow]*2-1);
    lrside[currentwindow] = wnewblock (0, 0, tilewidth[currentwindow]*2-1, windowmaxy[currentwindow]);
    /* Get blocks for the side (incoming) tiles */

    wsetscreen (scroll2[currentwindow]);              /* go to second one */
    wcls (0);                          /* clear it */
    
    windowwidth[currentwindow] = numx;
    windowheight[currentwindow] = numy;

    /* Set up window widths and heights */

    worldmaxx[currentwindow] = mapwidth[currentwindow]*tilewidth[currentwindow] - (numx*tilewidth[currentwindow]);
    worldmaxy[currentwindow] = mapheight[currentwindow]*tileheight[currentwindow] - (numy*tileheight[currentwindow]);
    /* Set maximum scrolling values, ie worldx cannot go past worldmaxx */

    /* Go through each sprite and store the widths and heights */
    for (wgtscrollv3 = 1; wgtscrollv3 < 1001; wgtscrollv3++)
        if (sprites[wgtscrollv3] != NULL)
        {
            spritewidth[wgtscrollv3] = wgetblockwidth (sprites[wgtscrollv3]);
            spriteheight[wgtscrollv3] = wgetblockheight (sprites[wgtscrollv3]);
        }

}


int wgetworldblock (int numx, int numy, wgtmap scrollmap)
/* Returns a tile from the location on a map */
{
    wgtmap tempmap;
    
    numx = numx / tilewidth[currentwindow];
    numy = numy / tileheight[currentwindow];
    /* Divide by 16 to convert world coords to map coords */

    tempmap = scrollmap;
    /* Make a ptr to the map */

    tempmap += numy*mapwidth[currentwindow] + numx;
    /* Move to the offset of the tile we want */

    return(*tempmap); /* Return the character at that position */
}

void wputworldblock (int numx, int numy, int bnum, wgtmap scrollmap,int currentwindow)
/* Puts a tile onto the map at the given world coordinates.
   This routine is very slow since the tiles must be put onto the raw data
   screen, which means 4 putblocks are required, one for each quadrant.
   A better way of making animating tiles is to make objects overtop
   a tile and animate that instead. This way you can have two layers
   for the tile as well. */
{
    int changeb;
    wgtmap tempmap;
    block *blocks;

    blocks=scrolltiles[currentwindow];
    numx = numx / tilewidth[currentwindow];
    numy = numy / tileheight[currentwindow];
    /* Convert world coords to map coords by dividing by 16 */

    tempmap = scrollmap;
    tempmap += numy*mapwidth[currentwindow] + numx;
    *tempmap = bnum;
    /* Get a character from the map */
	
    
    numx = (numx * tilewidth[currentwindow]) - worldx[currentwindow];
    numy = (numy * tileheight[currentwindow]) - worldy[currentwindow];
    /* Get the top left corner of the block to put */
    

    wsetscreen (scroll2[currentwindow]);

    /* Now put the block in each of the four quadrants using clipping */
    tx = scrllx[currentwindow]; ty = scrlly[currentwindow]; bx = windowmaxx[currentwindow]; by = windowmaxy[currentwindow];
    wputtile (numx + scrllx[currentwindow], numy + scrlly[currentwindow], blocks[bnum], scroll2[currentwindow]);
    
    tx = windowminx[currentwindow]; ty = scrlly[currentwindow]; bx = scrllx[currentwindow] - 1; by = windowmaxy[currentwindow];
    wputtile (numx - (windowmaxx[currentwindow] - scrllx[currentwindow]) - 1, numy + scrlly[currentwindow], blocks[bnum], scroll2[currentwindow]);
    
    tx = scrllx[currentwindow]; ty = windowminy[currentwindow]; bx = windowmaxx[currentwindow]; by = scrlly[currentwindow] - 1;
    wputtile (numx + scrllx[currentwindow], numy - (windowmaxy[currentwindow] - scrlly[currentwindow]) - 1, blocks[bnum], scroll2[currentwindow]);
    
    tx = windowminx[currentwindow]; ty = windowminy[currentwindow]; bx = scrllx[currentwindow] - 1; by = scrlly[currentwindow] - 1;
    wputtile (numx - (windowmaxx[currentwindow] - scrllx[currentwindow]) - 1, numy - (windowmaxy[currentwindow] - scrlly[currentwindow]) - 1, blocks[bnum], scroll2[currentwindow]);
    
    redrawside[currentwindow] = 1;
    /* Make sure we redraw the side blocks since the tiles have changed */
	
}


void wshowobjects (void)
/* Display the objects on top of the assembled data screen */
{
    wclip (0, 0, windowmaxx[currentwindow], windowmaxy[currentwindow]);

    /* Go through all sprites, up to numsprites, and draw them */
    for (wgtscrollv3 = 0; wgtscrollv3 <= numsprites; wgtscrollv3++)
    {
        if ((wobject[wgtscrollv3].on == 1) &
		(sprites[wobject[wgtscrollv3].num] != NULL)) 
	/* Make sure the object is on, and the block isn't null */

        {
	    if ((wobject[wgtscrollv3].x < worldx[currentwindow] + windowmaxx[currentwindow]) &
		    (wobject[wgtscrollv3].y < worldy[currentwindow] + windowmaxy[currentwindow]) &
		    (wobject[wgtscrollv3].x + spritewidth[wobject[wgtscrollv3].num] > worldx[currentwindow]) &
		    (wobject[wgtscrollv3].y + spriteheight[wobject[wgtscrollv3].num] > worldy[currentwindow]))
	    /* Check to see if the object is on the screen */

		wputanywhere (wobject[wgtscrollv3].x - worldx[currentwindow], wobject[wgtscrollv3].y - worldy[currentwindow], sprites[wobject[wgtscrollv3].num], scroll1[currentwindow], 1);
	    /* Put the block onto the scroll1 block */
	}
    }
}



void wfreemap (wgtmap ptr)
/* Frees memory from a map file */
{
    farfree (ptr);
}


void wendscroll (int currentwindow)
/* Closes up scrolling, and frees virtual screens */
{
    wfreeblock (scroll1[currentwindow]);
    wfreeblock (scroll2[currentwindow]);
    wfreeblock (udside[currentwindow]);
    wfreeblock (lrside[currentwindow]);
    wclip (0, 0, 319, 199);
    wnormscreen ();
}



void wcopyscroll (int x1, int y1,int currentwindow)
/* Copies the scrolling window to the visual screen */
{
    wshowscroller (0, 0, windowmaxx[currentwindow], windowmaxy[currentwindow], scroll1[currentwindow], x1, y1, NULL);
}


/***================================================================***
*             Put a tile onto the screen			      *
****================================================================***/
void wputtile (int x, int y, block pr, block dstblock)
{
    int offset,width,height,display,maxy;
    int ctr;
    block dst;
    unsigned char t;
    width = tilewidth[currentwindow];
    height = tileheight[currentwindow];
    pr += 4;
    wclip (0, 0, wgetblockwidth (dstblock) - 1, wgetblockheight (dstblock) - 1);
    
    if (x + width > bx) display = (bx + 1) - x;
    else display = width;
    maxy = y + height - 1;
    if (maxy > by) maxy = by;
    if (y < ty) 
    {
	pr += (ty - y) * tilewidth[currentwindow];   y = ty;
    }                                  /* clip y */
    if (x < tx) 
    {
        pr += tx - x; display -= tx - x;   x = tx;  
    }                                  /* clip x */
    if (display > 0)
    {
	dst = dstblock + 4 + y*screenwidth[currentwindow] + x;
	for (ctr = y; ctr <= maxy; ctr++) 
	{
	    memcpy (dst, pr, display);
	    dst += screenwidth[currentwindow];
	    pr += tilewidth[currentwindow];
        }
    }
}


/***================================================================***
*             Put a block using either copy or xray modes             *
*	      but allows putting onto any size block, not just	      *
*	      full screen 320x200 blocks (used for side tiles	      *
*	      and main scrolling screens			      *
****================================================================***/
void wputtileanywhere (int x, int y, block pr, block sideptr)
{
    int offset,width,height,display,maxy;
    int dwidth,dheight;
    int ctr,toadd;
    block dst;
    unsigned char t;
    
    dst = pr;
    t = *dst; dst++; width = t;
    t = *dst; dst++; width += t*256;   /* get width and height of block */
    t = *dst; dst++; height = t;       /* of block to put anywhere */
    t = *dst; dst++; height += t*256;
    
    dst = sideptr;
    t = *dst; dst++; dwidth = t;
    t = *dst; dst++; dwidth += t*256;  /* get width and height of block */
    t = *dst; dst++; dheight = t;      /* that is being put onto (dest) */
    t = *dst; dst++; dheight += t*256;
    sideptr+=4;

    pr += 4;

    if (x + width > bx) display = (bx + 1) - x;
    else display = width;
    maxy = y + height - 1;
    if (maxy > dheight) maxy = dheight;
    if (maxy > by) maxy = by;
    if (y < ty)
    {
	pr += (ty - y) *width;   y = ty;  
    }                                  /* clip y */
    if (x < tx) 
    {
        pr += tx - x; display -= tx - x;   x = tx;  
    }                                  /* clip x */
   if (display > 0)
    {
	toadd=dwidth-display;
	if (toadd<0) toadd=0;
	display>>=2;


        asm push ds;
        asm cld;
	asm lds si,pr;
	asm les di,sideptr;
	for (ctr = 0; ctr < height; ctr++)
	{
	    asm .386
	    asm mov cx,display;
	    asm rep movsd;
	    asm add di,toadd;
	    asm inc si;
	}
	asm pop ds;
    }
}


/***================================================================***
*             Put a block using either copy or xray modes             *
*	      but allows putting onto any size block, not just	      *
*	      full screen 320x200 blocks (used for side tiles	      *
*	      and main scrolling screens			      *
****================================================================***/
void wputanywhere (int x, int y, block pr, block sideptr, int method)
{
    int offset,width,height,display,maxy;
    int dwidth,dheight;
    int ctr;
    block dst;
    unsigned char t;
    
    dst = pr;
    t = *dst; dst++; width = t;
    t = *dst; dst++; width += t*256;   /* get width and height of block */
    t = *dst; dst++; height = t;       /* of block to put anywhere */
    t = *dst; dst++; height += t*256;
    
    dst = sideptr;
    t = *dst; dst++; dwidth = t;
    t = *dst; dst++; dwidth += t*256;  /* get width and height of block */
    t = *dst; dst++; dheight = t;      /* that is being put onto (dest) */
    t = *dst; dst++; dheight += t*256;
    sideptr+=4;

    pr += 4;
    wclip (0, 0, dwidth - 1, dheight - 1);
    
    
    if (x + width > bx) display = (bx + 1) - x;
    else display = width;
    maxy = y + height - 1;
    if (maxy > dheight) maxy = dheight;
    if (maxy > by) maxy = by;
    if (y < ty)
    {
	pr += (ty - y) *width;   y = ty;  
    }                                  /* clip y */
    if (x < tx) 
    {
        pr += tx - x; display -= tx - x;   x = tx;  
    }                                  /* clip x */
    if ((method == 0) & (display > 0))
    {
	dst = sideptr + y*dwidth + x;
        for (ctr = y; ctr <= maxy; ctr++) 
        {
            memcpy (dst, pr, display);
            dst += dwidth;
            pr += width;
        }
    }
    else if ((method == 1) & (display > 0))
    {
	dst = sideptr + y*dwidth + x;
        asm push ds;
        asm cld;
        for (ctr = y; ctr <= maxy; ctr++)
        {
            asm mov cx,display;
            asm lds si,pr;
            asm les di,dst;
            xloop:
            ;
            asm lodsb;
	    asm or al,al;
	    asm jz ok;
            asm mov es:di,al;
            ok:
            ;
            asm inc di;
            asm loop xloop
                dst += dwidth;
            pr += width;
        }
        asm pop ds;
    }
}



/***================================================================***
*             Put a block using either copy or xray modes             *
*	      but allows putting onto any size block, not just	      *
*	      full screen 320x200 blocks (used for side tiles	      *
*	      and main scrolling screens			      *
****================================================================***/
void wcopyanywhere (int x, int y, int x2, int y2, block pr, int dsx,int dsy,block to)
{
    int width,height,display,maxy;
    int xtra=0;
    register int ctr;
    int dwidth,dheight;                        /* temp window width */
    int swidth;                        /* temp source width */

    swidth = wgetblockwidth (pr); /* Get width of block */

     dwidth = wgetblockwidth (to); /* Set up dest width */
     dheight = wgetblockheight(to); /* Set up dest width */
    if (x2 < 0) x2 = 0;		       /* Check dimensions */
    if (x < 0) x = 0;
    if (y2 < 0) y2 = 0;
    if (y < 0) y = 0;

    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    /* Get width and height */

    if (x + width > dwidth) display = (dwidth) - x;
      else display = width;
    maxy = y + height - 1;
    if (maxy > dheight) maxy = dheight;
    if (maxy > by) maxy = by;
    if (y < 0)
    {
	pr += (0 - y) *width;   y = 0;
    }                                  /* clip y */
    if (x < 0)
    {
	pr += 0 - x; display -= 0 - x;   x = 0;
    }                                  /* clip x */


    display = width; /* Number of pixels to copy */
    if ((height > 0) && (display > 0))
    {
	    pr += 4; 			   /* Skip width and height integers */
	    to += 4;	/* Skip over width and height integers */

	pr += x + (y*swidth);     /* Move to correct offsets */
	to += dsx + (dsy*dwidth);

	xtra=display % 2;
	display=display/2;
	asm push ds;
	asm cld;
	for (ctr = y; ctr <=maxy; ctr++)
	{
	   if (display>0)
	    {
	    asm lds si,pr;
	    asm les di,to;
	    asm mov cx,display;
	    xloop:
	    ;
	    asm lodsb;
	    asm or al,al
	    asm jz ok;
            asm mov es:di,al;
            ok:
            ;
	    asm inc di;
	    asm or ah,ah
	    asm jz ok2;
	    asm mov es:di,al;
	    ok2:
	    ;
	    asm inc di;
	    asm loop xloop
	    }
	if (xtra)
	    {
	    asm lodsb;
	    asm or al,al
	    asm jz ok3;
            asm mov es:di,al;
	    ok3:
            ;
            asm inc di;
	    }
	    to += dwidth;
	    pr += swidth;
	}
	asm pop ds;
    }
}


  /*
/***================================================================***
*             Put a block using either copy or xray modes             *
*	      but allows putting onto any size block, not just	      *
*	      full screen 320x200 blocks (used for side tiles	      *
*	      and main scrolling screens			      *
****================================================================***/
void wcopyanywhere (int x, int y, int x2, int y2, block pr, int dsx,int dsy,block to)
{
    int width,height,display,maxy;
    register int ctr;
    int dwidth,dheight;                        /* temp window width */
    int swidth;                        /* temp source width */

    swidth = wgetblockwidth (pr); /* Get width of block */

     dwidth = wgetblockwidth (to); /* Set up dest width */
     dheight = wgetblockheight(to); /* Set up dest width */
    if (x2 < 0) x2 = 0;		       /* Check dimensions */
    if (x < 0) x = 0;
    if (y2 < 0) y2 = 0;
    if (y < 0) y = 0;

    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    /* Get width and height */

    if (x + width > dwidth) display = (dwidth) - x;
      else display = width;
    maxy = y + height - 1;
    if (maxy > dheight) maxy = dheight;
    if (maxy > by) maxy = by;
    if (y < 0)
    {
	pr += (0 - y) *width;   y = 0;
    }                                  /* clip y */
    if (x < 0)
    {
	pr += 0 - x; display -= 0 - x;   x = 0;
    }                                  /* clip x */


    display = width; /* Number of pixels to copy */
    if ((height > 0) && (display > 0))
    {
	    pr += 4; 			   /* Skip width and height integers */
	    to += 4;	/* Skip over width and height integers */

	pr += x + (y*swidth);     /* Move to correct offsets */
	to += dsx + (dsy*dwidth);

	asm push ds;
	asm cld;
	for (ctr = y; ctr <=maxy; ctr++)
	{
	    asm lds si,pr;
	    asm les di,to;
	    asm mov cx,display;
	    xloop:
	    ;
	    asm lodsb;
	    asm or al,al
	    asm jz ok;
            asm mov es:di,al;
            ok:
            ;
	    asm inc di;
	    asm loop xloop
	    to += dwidth;
	    pr += swidth;
	}
	asm pop ds;
    }
}

*/