#pragma inline
#include <dos.h>
#include <alloc.h>
#include <stdio.h>
#include <conio.h>
#include <fcntl.h>
#include <io.h>
#include <wgt.h>

/* Please note that the default tile when you first run the map maker is
   tile 201. This is remarked out in this version since we no longer write
   to 320x200 blocks. In order to do so, we would need to rewrite the bar 
   command to write with different widths.  I'm saving that for a future
   release in which I'll change all of the procedures, so you can draw on
   blocks of any width! The blocks are now only as big as you need them,
   to decrease memory usage.  If you have the default tile in any or your maps, you should
   scan through the map and replace any occurences of this tile with a new one,
   right after you load the map in.
*/

/**************************************************************************
 WordUp Graphics Toolkit source Copyright 1993 WordUp Software Productions
 Licensed for personal use only

 Written by:
			 Chris Egerter
			 94 Andover Drive
			 London Ontario, Canada
			 N6J 3X2
			 (519)472-3661

Belongs to library: WGT4SCR.LIB

***************************************************************************/

/*  		      Multidirectional Scrolling Library 

Naturally this is the most complicated set of routines in the toolkit.
I won't go into exact details throughout the code since I don't think you'll
want to touch much of this.  You probably won't be able to change anything
unless you understand all of the code perfectly, and even I don't know it.
It took me over two months to develop, and there are still bugs in it.

Here is the basic idea:

Let's begin with a pattern of tiles on the screen.

****************	Suppose we want to scroll the screen opposite
*T***T****H***HH 	to the left. Some scrollers simply shift the
*T***T****H*****	screen in the direction, and draw the new tiles
*TTTTT**HHHHH***	at the edges as needed. You would then need to
********H***H***	copy the result to another virtual page, put on
********H***H***	the sprites, and then finally copy it to the visual
   Frame   1		screen. This means there are 3 copy operations.

***************-	To the left is the screen shifted over by one tile.
T***T****H***HH-	WGT uses a faster method of copying, which only uses
T***T****H*****-	two copy operations.  Raw scroll data is stored on
TTTTT**HHHHH***-	block scroll2.  This screen is copied in 4 different
*******H***H***-	chunks over to different positions on block scroll1.
*******H***H***-	The sprites are then pasted on, and the whole window
    Frame 2    ^	is copied to the visual screen.
	       |
	       The dashes represent the new area which is scrolled on.

While this is a much more complicated method than the first, it requires
one less full screen copy.  When an area of the screen scrolls, notice how
most of the area does not change.  Only the area at the edges need to
be redrawn.  This is why redrawing all of the tiles on the screen does not
make sense.  Why redraw all of the tiles when only two of the edges change
at the most? (you can't scroll left and right or up and down at the same
time, only up left or down right, etc)

By keeping a coordinate placement and increasing the coordinate, we can
easily make an area scroll across the screen with two copy operations
The ^ shows where the placement is.  In each frame the area starting
at ^ to the end is copied to the first column in the result frame.
The columns before the ^ are copied to the column at the ^.
In other words, the screen is split into two sections at the placement,
and both are copied to the opposite sides of the screen.

Raw frames: (scroll2)
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
123456789   123456789   123456789   123456789   123456789
^            ^            ^            ^            ^
| 	     |    	---   	    ----  	-----
|	    |           |           |           |    
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
123456789   234567891   345678912   456789123   567891234
Result frames (scroll1)

As you can see, the screen appears to scroll left as the placement
moves right. The only problem is the scroll wraps around the side,
and infinitely scrolls left. To make new tiles scroll appear on the
right side, we just draw new ones onto the raw frame, at one column
before the placement. Hopefully you can follow this description, 
because it gets much more complicated than that...

Of course a game that scrolls by tiles at a time wouldn't be very
smooth. To scroll by one pixel at a time, you need to increase the
placement by one, and only draw on 1 column of the new tiles.
WGT does this by keeping a block with two tiles across (one column 
for the left side, and one for the right).  Every sixteen pixels that
you scroll, the sides are updated, and new tiles are drawn on these
temporary blocks.  This means only the new tiles on the side are drawn
one out of every 16 pixels scrolled.  This is obviously much faster
than drawing the tiles on every time, since most of the time needed
to draw a single frame is just using memcpy about 5-6 times. 

This is the basic idea of multidirectional scrolling the WGT way.
The above example only accounts for one direction. WGT also takes
care of the up and down scrolling as well, in the similar manner.
The only difference between the above example and what is used here, is
the fact that the screen is copied in four sections, not two.
(Because we're combining up/down and left/right)

Sprites do not need to have their background saved, since in each frame
we are copying the whole scroll window over from scroll2. The sprites
can be drawn on scroll1, and will be erased the next time the window
is updated.


WARNING! This concludes the beginning explanation.  It gets much harder
to understand after this point! I am not responsible if you fry your
brain trying to figure out what everything does here...
You can modify it if you want, but the code is very touchy, and you
will probably get graphics messed up all over the screen if you change
something critical. (Believe me, I've tried it!)

*/

// users should know about these
block scroll2[2];  // raw scroll data - looks messed up if viewed normally
block scroll1[2];	// assembled scroll data, copied from scroll2, but in different
		// locations. The sprites are drawn overtop onto this block
block udside[2];	// A block with dimensions windowmaxx+1 by 32
		// The extra 1 in windowmaxx+1 is because my math isn't right
		// somewhere....

		// Contains the top and bottom rows of tiles which appear
		// next according to the worldy position.
		// Tile are organized like this :  123456789...  (top)
		//				   123456789...  (bottom)

block lrside[2];// A block with dimensions 32 by windowmaxy+1
		// Contains the left and right columns of tiles which appear
		// next according to the worldx position.
		// Tile are organized like this :  11
		//				   22
		//				   33
		//                                 44
		//                                 .. 
		//			left-------^^------right

int mapwidth[2],mapheight[2]; // the width and height (in tiles) of the current map loaded


int spritewidth[1001],spriteheight[1001]; // width and height of sprites
					  // so we don't need to recalc all the time
int numsprites;				  // # of sprites to show with wshowobjects

typedef struct {		 // object structure
	char on;		 // sprite is turned on=1
	int x;			 // world x coordinate (max 320*16=5120)
	int y;			 // world y coordinate (max 200*16=3200)
	unsigned int num;	 // sprite # from sprites[1001] to show
	} scrollsprites;
scrollsprites wobject[1001];	 // make a 1000 (1..1000) objects

char tiletype[201];		 // tile attributes

int wgtparallax=0;		 // flag to use copyscreen or copyunder
int windowminx[2],windowminy[2],windowmaxx[2],windowmaxy[2]; // coords of scrolling area
				 // min coords are always 0 for now

int screenwidth[2];  		 // width of scrolling window
int worldx[2]={0,0},worldy[2]={0,0};           // world coordinates of top left corner of
				 // scrolling window

int worldmaxx[2]={320,320},worldmaxy[2]={320,320}; // world max x and y for window
				 // These are calculated so you can't scroll
				 // past the end of the world, depends on
				 // window size
int windowwidth[2],windowheight[2];    // Size of window (in tiles)

// A copyright statement to store in executables 
char *copyrscroll="WordUp Graphics Toolkit:Multidirectional Scrolling Library Copyright 1993 WordUp Software Productions";


// externals
extern block sprites[1001];			// sprite array
extern block blocks[1001];			// tile array
extern int kbdon[129];				// keyboard array (on/off)
// extern int kbdscanlist[129];			// used for old method
extern int tx,ty,bx,by;				// clipping vars
				// top x, top y, bottom x, bottom y

void wputanywhere (int, int, block, block, int);// puts a block onto another
		// block of ANY size, not limited to full screen 320x200 blocks
		// used for showing sprites onto scroll1

// private stuff
static int wgtscrollv3,wgtscrollv4; // misc math for scrolling
static int wgtscrollv5;		    // misc var for math routines
static int x_repeat,y_repeat; 	    // these are used to store the amount
	// to scroll if the side tile blocks need to be updated halfway
	// through the update. If the side tiles are already halfway copied,
	// and we're scrolling by 16 pixels, 8 pixels of the current tiles
	// must be copied, and then the new tiles must be redrawn and 8
	// more pixels copied.

static int orig_x_speed,orig_y_speed; // stores original speeds in case
			// the condition above with x_repeat occurs
static int redrawside[2];  // if wputworldblock is used, all of the side tiles
		// are redrawn, in case they need to be updated. This is
		// probably why animating tiles slow the scrolling down a lot...

static int origkey;	// original location in keyboard buffer
static int intkey;	// current location in keyboard buffer
static int cabsx,cabsy; // absolute value of the speedx and speedy vars

static int column_tile_loc[2]={0,0};
static int row_tile_loc[2]={0,0};
static int scrlly[2]={0,0},scrllx[2]={0,0};                // scrll is ctr for offset
static int upredraw[2],downredraw[2],leftredraw[2],rightredraw[2];
static int speedx,speedy;

static void wqcopyscreen (int, int, int, int, block ptr, int, int, block);
static void wcopyside (int, int, int, int, block ptr, int, int, block);
static void wcopyunder (int, int, int, int, block ptr, int, int, block);
static void scl (void);
static void scr (void);
static void scu (void);
static void scd (void);
void wputtile (int, int, block, block);
static void interrupt kbd (); /* interrupt prototype */
static void interrupt (*oldfunc)(); /* interrupt function pointer */


// funcs
void installkbd (void);
void winitscroll (int, int);
void wshowwindow (int, int, wgtmap);
void wscrollwindow (int, int, wgtmap);		// scroll the screen
void wshowobjects (void);
int  wgetworldblock (int, int, wgtmap);
void wputworldblock (int, int, int, wgtmap);
void wcopyscroll (int, int);
void wendscroll (void);
void uninstallkbd (void);
void wfreemap (wgtmap);
int soverlap (int, int);
int currentwindow=0;

wgtmap wloadmap (char *mapfile)
{
    //extern char tiletype[201];
    wgtmap ptr,orig;
    int ctr,numread,mapver;
    unsigned char conv;
    unsigned size,bytes;
    
    if (wgtlibrary == NULL)
    {
        if ((libf = fopen (mapfile, "rb")) == NULL)
            return NULL;
    }
    else 
    {
        if ((libf = fopen (wgtlibrary, "rb")) == NULL)
            return NULL;
        readheader ();
        findfile (mapfile);
        if (lresult == 1)
            fseek (libf, lfpos, SEEK_SET);
        if (checkpassword (password) == 0)
        {
            textmode (C80);
            printf ("Incorrect password");
            exit (1);
        }
    }
    
    if ((wgtlibrary != NULL) & (lresult == 0)) goto mapstop;
    
    mapver = getw (libf);
    if ((mapver != 8972) & (mapver != 8973))
    {
        textmode (C80);
	printf ("Invalid map header: %i\n", mapwidth[currentwindow]);
        exit (1);
    }
    if (mapver == 8972)
    {
	mapwidth[currentwindow] = fgetc (libf);
	mapheight[currentwindow] = fgetc (libf);
    }
    else
    {
	mapwidth[currentwindow] = getw (libf);
	mapheight[currentwindow] = getw (libf);
    }
    
    size = mapheight[currentwindow]*mapwidth[currentwindow];
    if (size < 65533) 
    {
        ptr = farmalloc (size);
        if (ptr == NULL) return NULL;
        orig = ptr;
        if (fread (ptr, size, 1, libf) < 1) 
        {
            printf ("Read Failed (MAP).\n");
            exit (1);
        }
        for (ctr = 1; ctr <= 200; ctr++)
            tiletype[ctr] = getc (libf);
        ctr = getw (libf);
        if (mapver == 8973)
            fread (&wobject[1], 7, ctr, libf);
        else 
        {
            for (numread = 1; numread <= ctr; numread++)
            {
                fread (&wobject[numread].on, 1, 1, libf);
                fread (&wobject[numread].x, 2, 1, libf);
                fread (&wobject[numread].y, 2, 1, libf);
                fread (&conv, 1, 1, libf);
                wobject[numread].num = (unsigned int)conv;
            }
        }
    }
    mapstop:
    ;
    fclose (libf);
    return orig;
}


void wsavemap (char *fn, wgtmap map)
{
    FILE *fp;
    wgtmap ptr,orig;
    int ctr,numread,mapver,ynum;
    unsigned char conv;
    unsigned size,bytes;
    
    fp = fopen (fn, "wb");
    putw (8973, fp);
    putw (mapwidth[currentwindow], fp);
    putw (mapheight[currentwindow], fp);
    
    size = mapheight[currentwindow]*mapwidth[currentwindow];
    fwrite (map, size, 1, fp);
    for (ctr = 1; ctr <= 200; ctr++)
        fputc (tiletype[ctr], fp);
    
    ynum = 0;
    for (numread = 1; numread < 1001; numread++)
        if (wobject[numread].num > 0) ynum = numread;
    putw (ynum, fp);
    if (ynum > 0) fwrite (&wobject[1], 7*ynum, 1, fp);
    
    fclose (fp);
}


int soverlap (int s1, int s2)
{
    int n1,n2;
    
    if ((wobject[s2].on & wobject[s1].on == 1))
    {
        n1 = wobject[s1].num;
        n2 = wobject[s2].num;
        if (( wobject[s2].x >= wobject[s1].x - spritewidth[n2] ) &&
                ( wobject[s2].x <= wobject[s1].x + spritewidth[n1] ) &&
                ( wobject[s2].y >= wobject[s1].y - spriteheight[n2] ) &&
                ( wobject[s2].y <= wobject[s1].y + spriteheight[n1] )) return 1;
    }
    return 0;
}

void wscrollwindow (int wsx, int wsy, wgtmap scrollmap)
{
    wgtmap tempmap;
    int q;
    
    speedx = wsx;
    speedy = wsy;
    abuf = MK_FP (FP_SEG (scroll2[currentwindow] + 4), FP_OFF (scroll2[currentwindow] + 4));
    
    if (speedx > 16) speedx = 16;
    else if (speedx < - 16) speedx = - 16;
    if (speedy > 16) speedy = 16;
    else if (speedy < - 16) speedy = - 16;
    
    if ((x_repeat) | (y_repeat))
    {
	speedx = x_repeat;
	speedy = y_repeat;
    }
    
    x_repeat = 0; orig_x_speed = speedx;
    y_repeat = 0; orig_y_speed = speedy;
    
    
    if (speedx)
    {
	upredraw[currentwindow] = 1; downredraw[currentwindow] = 1;
	
	if (worldx[currentwindow] + speedx < 0)
	    speedx += abs (worldx[currentwindow] + speedx);
	else if (worldx[currentwindow] + speedx > worldmaxx[currentwindow])
	    speedx -= abs (worldmaxx[currentwindow] - (worldx[currentwindow] + speedx));
	
	if ((column_tile_loc[currentwindow] + speedx > 16) && (column_tile_loc[currentwindow] != 16))
	{
	    speedx = 16 - column_tile_loc[currentwindow];
	    column_tile_loc[currentwindow] = 16;
	    x_repeat = orig_x_speed - speedx;
        }
	else if ((column_tile_loc[currentwindow] + speedx < 0) && (column_tile_loc[currentwindow] != 0))
	{
	    speedx = - column_tile_loc[currentwindow];
	    column_tile_loc[currentwindow] = 0;
	    x_repeat = orig_x_speed - speedx;
        }
        else  
        {
	    column_tile_loc[currentwindow] += speedx;
	    if (column_tile_loc[currentwindow] < 0)
            {
		column_tile_loc[currentwindow] += 16; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
	    }
	    else if (column_tile_loc[currentwindow] > 16)
	    {
		column_tile_loc[currentwindow] -= 16; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1; 
            }
        }
        
        
	worldx[currentwindow] += speedx;
	
	scrllx[currentwindow] += speedx;
	if ((speedx < 0) && (scrllx[currentwindow] < windowminx[currentwindow]))
	    scrllx[currentwindow] += windowmaxx[currentwindow] + 1;
	else if ((speedx > 0) && (scrllx[currentwindow] > windowmaxx[currentwindow]))
	    scrllx[currentwindow] -= windowmaxx[currentwindow] + 1;
        
    }
    
    if (speedy)
    {
	leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
	
	if (worldy[currentwindow] + speedy < 0)
	    speedy += abs (worldy[currentwindow] + speedy);
	else if (worldy[currentwindow] + speedy > worldmaxy[currentwindow])
	    speedy -= abs (worldmaxy[currentwindow] - (worldy[currentwindow] + speedy));
        
	if ((row_tile_loc[currentwindow] + speedy > 16) && (row_tile_loc[currentwindow] != 16))
	{
	    speedy = 16 - row_tile_loc[currentwindow];
	    row_tile_loc[currentwindow] = 16;
	    y_repeat = orig_y_speed - speedy;
        }
	else if ((row_tile_loc[currentwindow] + speedy < 0) && (row_tile_loc[currentwindow] != 0))
        {
	    speedy = - row_tile_loc[currentwindow];
	    row_tile_loc[currentwindow] = 0;
	    y_repeat = orig_y_speed - speedy;
        }
        else
        {
	    row_tile_loc[currentwindow] += speedy;
	    if (row_tile_loc[currentwindow] < 0)
	    {
		row_tile_loc[currentwindow] += 16; upredraw [currentwindow]= 1; downredraw[currentwindow] = 1; 
	    }
	    else if (row_tile_loc[currentwindow] > 16)
	    {
		row_tile_loc[currentwindow] -= 16; upredraw[currentwindow] = 1; downredraw[currentwindow] = 1; 
            }
        }
        
        
	worldy[currentwindow] += speedy;
	
	scrlly[currentwindow] += speedy;
	if ((speedy < 0) && (scrlly[currentwindow] < windowminy[currentwindow]))
	    scrlly[currentwindow] += windowmaxy[currentwindow] + 1;
	else if ((speedy > 0) && (scrlly[currentwindow] > windowmaxy[currentwindow]))
	    scrlly[currentwindow] -= windowmaxy[currentwindow] + 1;
        
    }
    
    
    
    if (((speedx > 0) && (rightredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    {
	if (redrawside[currentwindow] == 0)rightredraw[currentwindow] = 0;
	tx = 16; ty = 0; bx = 35; by = windowmaxy[currentwindow];
	wgtscrollv4 = worldy[currentwindow] % 16;
	if (wgtscrollv4 != 0)
	    wgtscrollv4 = 0 - wgtscrollv4;
	tempmap = scrollmap;
	tempmap += ( ((worldy[currentwindow]) >> 4)*mapwidth[currentwindow] + (((worldx[currentwindow] - 1) >> 4) + windowwidth[currentwindow]));
	for (q = 0; q <= windowheight[currentwindow] + 1; q++)
	{
	    wgtscrollv3 = *tempmap;
	    //  if (wgtscrollv3==201)
	    //     { wsetcolor(0); wbar(19,wgtscrollv4,34,wgtscrollv4+15); }
	    wputanywhere (16, wgtscrollv4, blocks[wgtscrollv3], lrside[currentwindow], 0);
	    wgtscrollv4 += 16;
	    tempmap += mapwidth[currentwindow];
	}
	//wclip(0,0,319,199);
	tx = 0; ty = 0; bx = 319; by = 199;
	//drawright();
    }
    
    if (((speedx < 0) && (leftredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    {
	if (redrawside [currentwindow]== 0)leftredraw[currentwindow] = 0;
	//wclip(windowmaxx+3,0,windowmaxx+18,windowmaxy);
	tx = 0; ty = 0; bx = 16; by = windowmaxy[currentwindow];
	wgtscrollv4 = worldy[currentwindow] % 16;
	if (wgtscrollv4 != 0)
	    wgtscrollv4 = 0 - wgtscrollv4;
	tempmap = scrollmap;
	tempmap += ( ((worldy[currentwindow]) >> 4)*mapwidth[currentwindow] + (worldx[currentwindow] >> 4));
	for (q = 0; q <= windowheight[currentwindow] + 1; q++)
	{
	    //  wgtscrollv3=map[(worldy[currentwindow]+(q<<4))>> 4][worldx[currentwindow] >> 4];
            wgtscrollv3 = *tempmap;
            //  if (wgtscrollv3==201)
            //     { wsetcolor(0); wbar(3,wgtscrollv4,18,wgtscrollv4+15); }
	    wputanywhere (0, wgtscrollv4, blocks[wgtscrollv3], lrside[currentwindow], 0);
	    wgtscrollv4 += 16;
	    tempmap += mapwidth[currentwindow];
	}
	//wclip(0,0,319,199);
	tx = 0; ty = 0; bx = 319; by = 199;
	// drawleft();
    }
    
    if (((speedy > 0) && (downredraw[currentwindow] != 0)) | (redrawside[currentwindow] == 1))
    {
	if (redrawside[currentwindow] == 0) downredraw[currentwindow] = 0;
	//wclip(0,windowmaxy+19,windowmaxx,windowmaxy+35);
	tx = 0; ty = 16; bx = windowmaxx[currentwindow]; by = 35;
	wgtscrollv4 = worldx[currentwindow] % 16;
	if (wgtscrollv4 != 0)
	    wgtscrollv4 = 0 - wgtscrollv4;
	tempmap = scrollmap;
	tempmap += ( (((worldy[currentwindow] - 1) >> 4) + windowheight[currentwindow])*mapwidth[currentwindow] + ((worldx[currentwindow]) >> 4));
	for (q = 0; q <= windowwidth[currentwindow] + 1; q++)
	{
	    //  wgtscrollv3=map[((worldy[currentwindow]-1)>>4)+windowheight[currentwindow]][(worldx[currentwindow]+(q<<4))>>4];
	    wgtscrollv3 = *tempmap;
	    
	    //  if (wgtscrollv3==201)
	    //     { wsetcolor(0); wbar(wgtscrollv4,windowmaxy+19,wgtscrollv4+15,windowmaxy+34); }
	    wputanywhere (wgtscrollv4, 16, blocks[wgtscrollv3], udside[currentwindow], 0);
	    wgtscrollv4 += 16;
	    tempmap++;
        }
        //wclip(0,0,319,199);
        tx = 0; ty = 0; bx = 319; by = 199;
        // drawdown();
    }
    
    if (((speedy < 0) && (upredraw[currentwindow])) | (redrawside[currentwindow] == 1))
    {
	if (redrawside[currentwindow] == 0)upredraw[currentwindow] = 0;
	//wclip(0,windowmaxy+3,windowmaxx,windowmaxy+18);
	tx = 0; ty = 0; bx = windowmaxx[currentwindow]; by = 16;
	wgtscrollv4 = worldx[currentwindow] % 16;
	if (wgtscrollv4 != 0)
	    wgtscrollv4 = 0 - wgtscrollv4;
	tempmap = scrollmap;
	tempmap += ( ((worldy[currentwindow] >> 4)*mapwidth[currentwindow]) + ((worldx[currentwindow]) >> 4));
	for (q = 0; q <= windowwidth[currentwindow] + 1; q++)
	{
            wgtscrollv3 = *tempmap;
            //  if (wgtscrollv3==201)
            //     { wsetcolor(0); wbar(wgtscrollv4,windowmaxy+3,wgtscrollv4+15,windowmaxy+18); }
	    wputanywhere (wgtscrollv4, 0, blocks[wgtscrollv3], udside[currentwindow], 0);
	    wgtscrollv4 += 16;
	    tempmap++;
	}
	//wclip(0,0,319,199);
	tx = 0; ty = 0; bx = 319; by = 199;
	//drawup();
	
    }
    
    wsetscreen (scroll2[currentwindow]);
    cabsx = abs (speedx);
    cabsy = abs (speedy);
    
    if (speedx > 0)
        scr ();
    else if (speedx < 0)
        scl ();
    if (speedy > 0)
        scd ();
    else if (speedy < 0)
        scu ();
    
    
    tx = 0; ty = 0; bx = 319; by = 199;
    if ((x_repeat == 0) && (y_repeat == 0))
    {
	wgtscrollv3 = windowmaxx[currentwindow] - scrllx[currentwindow] + 1;
	wgtscrollv4 = scrllx[currentwindow] - 1;
        
//        if (wgtparallax == 0)
//        {
	    wqcopyscreen (windowminx[currentwindow], windowminy[currentwindow], wgtscrollv4, scrlly[currentwindow], scroll2[currentwindow], wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	    // 1
	    wqcopyscreen (scrllx[currentwindow], windowminy[currentwindow], windowmaxx[currentwindow], scrlly[currentwindow] - 1, scroll2[currentwindow], windowminx[currentwindow], windowmaxy[currentwindow] - scrlly[currentwindow] + 1, scroll1[currentwindow]);
	    // 2
	    wqcopyscreen (windowminx[currentwindow], scrlly[currentwindow], wgtscrollv4, windowmaxy[currentwindow], scroll2[currentwindow], wgtscrollv3, windowminy[currentwindow], scroll1[currentwindow]);
	    // 3
	    wqcopyscreen (scrllx[currentwindow], scrlly[currentwindow], windowmaxx[currentwindow], windowmaxy[currentwindow], scroll2[currentwindow], windowminx[currentwindow], windowminy[currentwindow], scroll1[currentwindow]);
            // 4
  /*	}
	else 
	{    
            wcopyunder (windowminx, windowminy, wgtscrollv4, scrlly, scroll2, wgtscrollv3, windowmaxy - scrlly + 1, scroll1);
            // 1
            wcopyunder (scrllx, windowminy, windowmaxx, scrlly - 1, scroll2, windowminx, windowmaxy - scrlly + 1, scroll1);
            // 2
            wcopyunder (windowminx, scrlly, wgtscrollv4, windowmaxy, scroll2, wgtscrollv3, windowminy, scroll1);
            // 3
            wcopyunder (scrllx, scrlly, windowmaxx, windowmaxy, scroll2, windowminx, windowminy, scroll1);
            // 4
	}*/
    }
    
    if ((x_repeat != 0) | (y_repeat != 0))
        wscrollwindow (speedx, speedy, scrollmap);
    redrawside[currentwindow] = 0;
    
    /*wnormscreen();
    wputblock(0,0,udside,0);
    wputblock(0,0,lrside,0);*/
}

void scl (void)
{
    wgtscrollv3 = column_tile_loc[currentwindow];
    wgtscrollv4 = wgtscrollv3 + cabsx - 1;
    wgtscrollv5 = scrllx[currentwindow];
    if (wgtscrollv5 > windowmaxx[currentwindow])
	wgtscrollv5 -= windowmaxx[currentwindow] + 1;
    wcopyside (wgtscrollv3, 0, wgtscrollv4, windowmaxy[currentwindow] - scrlly[currentwindow], lrside[currentwindow], wgtscrollv5, scrlly[currentwindow], scroll2[currentwindow]);
    if (scrlly[currentwindow] > 0)
	wcopyside (wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, wgtscrollv4, windowmaxy[currentwindow], lrside[currentwindow], wgtscrollv5, 0, scroll2[currentwindow]);
}

void scr (void)
{
    wgtscrollv3 = 16 + column_tile_loc[currentwindow] - cabsx;
    wgtscrollv4 = wgtscrollv3 + cabsx - 1;
    wgtscrollv5 = scrllx[currentwindow] - cabsx;
    if (wgtscrollv5 < 0)
	wgtscrollv5 += windowmaxx[currentwindow] + 1;
    wcopyside (wgtscrollv3, 0, wgtscrollv4, windowmaxy[currentwindow] - scrlly[currentwindow], lrside[currentwindow], wgtscrollv5, scrlly[currentwindow], scroll2[currentwindow]);
    if (scrlly[currentwindow]> 0)
	wcopyside (wgtscrollv3, windowmaxy[currentwindow] - scrlly[currentwindow] + 1, wgtscrollv4, windowmaxy[currentwindow], lrside[currentwindow], wgtscrollv5, 0, scroll2[currentwindow]);
}

void scu (void)
{
    wgtscrollv3 = row_tile_loc[currentwindow];
    wgtscrollv4 = wgtscrollv3 + cabsy - 1;
    wgtscrollv5 = scrlly[currentwindow] - cabsy;
    wgtscrollv5 = scrlly[currentwindow];
    if (wgtscrollv5 > windowmaxy[currentwindow])
	wgtscrollv5 -= windowmaxy[currentwindow] + 1;
    wcopyside (0, wgtscrollv3, windowmaxx[currentwindow] - scrllx[currentwindow], wgtscrollv4, udside[currentwindow], scrllx[currentwindow], wgtscrollv5, scroll2[currentwindow]);
    if (scrllx[currentwindow] > 0)
	wcopyside (windowmaxx[currentwindow] - scrllx[currentwindow] + 1, wgtscrollv3, windowmaxx[currentwindow], wgtscrollv4, udside[currentwindow], 0, wgtscrollv5, scroll2[currentwindow]);
}

void scd (void)
{
    wgtscrollv3 = 16 + row_tile_loc[currentwindow] - cabsy;
    wgtscrollv4 = wgtscrollv3 + cabsy - 1;
    wgtscrollv5 = scrlly[currentwindow] - cabsy;
    if (wgtscrollv5 < 0)
	wgtscrollv5 += windowmaxy[currentwindow] + 1;
    wcopyside (0, wgtscrollv3, windowmaxx[currentwindow] - scrllx[currentwindow], wgtscrollv4, udside[currentwindow], scrllx[currentwindow], wgtscrollv5, scroll2[currentwindow]);
    if (scrllx[currentwindow] > 0)
	wcopyside (windowmaxx[currentwindow] - scrllx[currentwindow] + 1, wgtscrollv3, windowmaxx[currentwindow], wgtscrollv4, udside[currentwindow], 0, wgtscrollv5, scroll2[currentwindow]);
}

// Copyright 1992 Chris Egerter

void interrupt kbd (void)
{
    int i,j;
    
    disable ();
    i = inportb (0x60);
    j = 0;
    kbdon[i % 128] = (i < 128);
    
    // Old method...
    /*while ((kbdscanlist[j] !=0) & (j<100))
    {
    if (i==kbdscanlist[j])
    kbdon[j]=1;
    if (i==kbdscanlist[j]+128)
    kbdon[j]=0;
    j++;
    };*/
    (* oldfunc)();
    /* call old interrupt 9 handler            */
    pokeb (0x000, 0x041C, intkey);
    enable ();
}

void installkbd ()
{
    oldfunc = getvect (0x09);
    setvect (0x09, kbd);
    intkey = peekb (0x000, 0x041A);
    origkey = intkey;
    intkey += 2;
    if (intkey > 58)
        intkey = 30;
}


void uninstallkbd ()
{
    setvect (0x09, oldfunc);
    pokeb (0x000, 0x041A, origkey);
    delay (100); while (kbhit ()) getch ();
}


/***================================================================***
*                 Copy part of a screen to another                    *
****================================================================***/
void wcopyside (int x, int y, int x2, int y2, block ptr, int desx, int desy, block to)
{
    int width,height,display,maxy;
    int dwidth;
    register int ctr;
    dwidth = wgetblockwidth (ptr);
    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    display = width;
    if ((height > 0) && (display > 0))
    {
        ptr += 4 + x + (y*dwidth);
        to += 4;
	to += desx + (desy*screenwidth[currentwindow]);
	for (ctr = 0; ctr < height; ctr++) 
	{
	    memcpy (to, ptr, display);
	    ptr += dwidth;
	    to += screenwidth[currentwindow];
	}
    }
}
/***================================================================***
*                 Copy part of a screen to another                    *
****================================================================***/
void wqcopyscreen (int x, int y, int x2, int y2, block ptr, int desx, int desy, block to)
{
    int width,height,display,maxy;
    register int ctr;
    int dwidth;                        // temp window width
    int swidth;                        // temp source width
    if (x2 < 0) x2 = 0;
    if (x < 0) x = 0;
    if (y2 < 0) y2 = 0;
    if (y < 0) y = 0;
    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    
    display = width;
    if ((height > 0) && (display > 0))
    {
        if (ptr == NULL)
        {
            ptr = MK_FP (0xA000, 0x0000);
            swidth = 320;
        }
        else
        {
            swidth = wgetblockwidth (ptr);
            ptr += 4;
        }
        if (to == NULL)
        {
            to = MK_FP (0xA000, 0x0000);
            dwidth = 320;
        }
        else
        {
            dwidth = wgetblockwidth (to);
            to += 4;
        }
        ptr += x + (y*swidth);
        to += desx + (desy*dwidth);
        for (ctr = 0; ctr < height; ctr++) 
        {
            memcpy (to, ptr, display);
            ptr += swidth;
            to += dwidth;
        }
    }
}

/***================================================================***
*                 Copy part of a screen to another                    *
****================================================================***/
void wcopyunder (int x, int y, int x2, int y2, block rptr, int desx, int desy, block to)
{
    int width,height,display,maxy,tem;
    int ctr;
    width = abs (x - x2) + 1;
    height = abs (y - y2) + 1;
    display = width;
    if ((height > 0) && (display > 0))
    {
        if (rptr == NULL)
            rptr = MK_FP (0xA000, 0x0000);
        else
            rptr += 4;
        rptr += x + (y*320);
        if (to == NULL)
            to = MK_FP (0xA000, 0x0000);
        else
            to += 4;
        to += desx + (desy*320);
        asm push ds;
        tem = 320 - display;
        asm cld;
        asm lds si,rptr;
        asm les di,to;
        for (ctr = 0; ctr < height; ctr++) 
        {
            asm mov cx,display;
            xloop:
            ;
            //   asm lodsb;
            asm mov al, [si]           // get next pixel value into al
                asm       or al,al;
            asm     jz ok;
            asm      mov es:[di],al;
            ok:
            ;
            asm inc di;
            asm inc si;
	    asm loop xloop
                asm add di,tem
                asm add si,tem
                //   rptr+=320;
                //   to+=320;
                }
        asm pop ds;
    }
    
}


void wshowwindow (int numx, int numy, wgtmap scrollmap)
{
    int i,j;
    wgtmap tempmap;
    
    if ((numx*16 > worldmaxx[currentwindow]) | (numy*16 > worldmaxy[currentwindow]) | (numx < 0) | (numy < 0))
    {
        textmode (C80);
        printf ("\nError: Coordinates outside world in showwindow.");
        printf ("Coordinates used: %i,  %i\n", numx, numy);
	printf ("Max Coordinates:  %i,  %i\n", worldmaxx[currentwindow], worldmaxy[currentwindow]);
	exit (1);
    }
    
    wsetscreen (scroll2[currentwindow]);
    column_tile_loc[currentwindow] = 0; row_tile_loc[currentwindow] = 0;
    scrlly[currentwindow] = 0;scrllx[currentwindow] = 0;             // scrll is ctr for offset
    upredraw[currentwindow] = 1; downredraw[currentwindow] = 1; leftredraw[currentwindow] = 1; rightredraw[currentwindow] = 1;
    wclip (0, 0, 319, 199);
    for (i = 0; i < windowwidth[currentwindow]; i++)
	for (j = 0; j < windowheight[currentwindow]; j++)
    {
        tempmap = scrollmap;
	tempmap += (j + numy)*mapwidth[currentwindow] + (i + numx);
	wgtscrollv3 = *tempmap;
	if (wgtscrollv3 == 201)
	{
	    wsetcolor (0); wbar (windowmaxx[currentwindow] + 19, wgtscrollv4, windowmaxx[currentwindow] + 34, wgtscrollv4 + 15); 
	}
	else wputtile (i*16, j*16, blocks[wgtscrollv3], scroll2[currentwindow]);
    }
    worldx[currentwindow] = numx*16;
    worldy[currentwindow] = numy*16;
}


void winitscroll (int numx, int numy)
{
    if ((numx > 20) | (numy > 12) | (numx < 2) | (numy < 2))
    {
        textmode (C80);
        printf ("\nInvalid scroll window in call to initscroll!");
        exit (1);
    }
    
    windowminx[currentwindow] = 0; windowminy[currentwindow] = 0; windowmaxx[currentwindow] = numx*16 - 1; windowmaxy[currentwindow] = numy*16 - 1;
    // coords of scrolling area
    speedx = 0; speedy = 0;
    screenwidth[currentwindow] = windowmaxx[currentwindow] + 2;
    
    scroll1[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow] + 1, windowmaxy[currentwindow] + 1);
    // get two virtual screens
    scroll2[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow] + 1, windowmaxy[currentwindow] + 1);
    udside[currentwindow] = wnewblock (0, 0, windowmaxx[currentwindow], 31);
    lrside[currentwindow] = wnewblock (0, 0, 31, windowmaxy[currentwindow]);
    wsetscreen (scroll2[currentwindow]);              // go to second one
    wcls (0);                          // clear it
    
    windowwidth[currentwindow] = numx;
    windowheight[currentwindow] = numy;
    worldmaxx[currentwindow] = mapwidth[currentwindow]*16 - (numx*16); worldmaxy[currentwindow] = mapheight[currentwindow]*16 - (numy*16);
    for (wgtscrollv3 = 1; wgtscrollv3 < 1001; wgtscrollv3++)
        if (sprites[wgtscrollv3] != NULL)
        {
            spritewidth[wgtscrollv3] = wgetblockwidth (sprites[wgtscrollv3]);
            spriteheight[wgtscrollv3] = wgetblockheight (sprites[wgtscrollv3]);
        }
}


int wgetworldblock (int numx, int numy, wgtmap scrollmap)
{
    wgtmap tempmap;
    
    numx = numx >> 4;
    numy = numy >> 4;
    tempmap = scrollmap;
    tempmap += numy*mapwidth[currentwindow] + numx;
    return(*tempmap);
}

void wputworldblock (int numx, int numy, int bnum, wgtmap scrollmap)
{
    int changeb;
    wgtmap tempmap;
    
    numx = numx >> 4;
    numy = numy >> 4;
    tempmap = scrollmap;
    tempmap += numy*mapwidth[currentwindow] + numx;
    *tempmap = bnum;
    
    
    numx = (numx << 4) - worldx[currentwindow];
    numy = (numy << 4) - worldy[currentwindow];
    
    
    wsetscreen (scroll2[currentwindow]);
    //wclip(scrllx,scrlly,windowmaxx,windowmaxy);// 4
    tx = scrllx[currentwindow]; ty = scrlly[currentwindow]; bx = windowmaxx[currentwindow]; by = windowmaxy[currentwindow];
    wputtile (numx + scrllx[currentwindow], numy + scrlly[currentwindow], blocks[bnum], scroll2[currentwindow]);
    
    //wclip(windowminx,scrlly,scrllx-1,windowmaxy);// 3
    tx = windowminx[currentwindow]; ty = scrlly[currentwindow]; bx = scrllx[currentwindow] - 1; by = windowmaxy[currentwindow];
    wputtile (numx - (windowmaxx[currentwindow] - scrllx[currentwindow]) - 1, numy + scrlly[currentwindow], blocks[bnum], scroll2[currentwindow]);
    
    //wclip(scrllx,windowminy,windowmaxx,scrlly-1);// 2
    tx = scrllx[currentwindow]; ty = windowminy[currentwindow]; bx = windowmaxx[currentwindow]; by = scrlly[currentwindow] - 1;
    wputtile (numx + scrllx[currentwindow], numy - (windowmaxy[currentwindow] - scrlly[currentwindow]) - 1, blocks[bnum], scroll2[currentwindow]);
    
    //wclip(windowminx,windowminy,scrllx-1,scrlly-1);// 1
    tx = windowminx[currentwindow]; ty = windowminy[currentwindow]; bx = scrllx[currentwindow] - 1; by = scrlly[currentwindow] - 1;
    wputtile (numx - (windowmaxx[currentwindow] - scrllx[currentwindow]) - 1, numy - (windowmaxy[currentwindow] - scrlly[currentwindow]) - 1, blocks[bnum], scroll2[currentwindow]);
    
    redrawside[currentwindow] = 1;
    
}


void wshowobjects (void)
{
    wclip (0, 0, windowmaxx[currentwindow], windowmaxy[currentwindow]);
    for (wgtscrollv3 = 0; wgtscrollv3 <= numsprites; wgtscrollv3++)
    {
        if ((wobject[wgtscrollv3].on == 1) &
                (sprites[wobject[wgtscrollv3].num] != NULL))
        {
	    if ((wobject[wgtscrollv3].x < worldx[currentwindow] + windowmaxx[currentwindow]) &
		    (wobject[wgtscrollv3].y < worldy[currentwindow] + windowmaxy[currentwindow]) &
		    (wobject[wgtscrollv3].x + spritewidth[wobject[wgtscrollv3].num] > worldx[currentwindow]) &
		    (wobject[wgtscrollv3].y + spriteheight[wobject[wgtscrollv3].num] > worldy[currentwindow]))
		wputanywhere (wobject[wgtscrollv3].x - worldx[currentwindow], wobject[wgtscrollv3].y - worldy[currentwindow], sprites[wobject[wgtscrollv3].num], scroll1[currentwindow], 1);
	}
    }
}

void wfreemap (wgtmap ptr)
{
    if (ptr !=NULL)
      farfree (ptr);
}

void wendscroll ()
{
    wfreeblock (scroll1[currentwindow]);
    wfreeblock (scroll2[currentwindow]);
    wfreeblock (udside[currentwindow]);
    wfreeblock (lrside[currentwindow]);
    wclip (0, 0, 319, 199);
    wnormscreen ();
}

void wcopyscroll (int x1, int y1)
{
    wqcopyscreen (0, 0, windowmaxx[currentwindow], windowmaxy[currentwindow], scroll1[currentwindow], x1, y1, NULL);
}

/***================================================================***
*             Put a tile onto the screen			      *
****================================================================***/
void wputtile (int x, int y, block pr, block dstblock)
{
    int offset,width,height,display,maxy;
    int ctr;
    block dst;
    unsigned char t;
    width = 16;
    height = 16;
    pr += 4;
    wclip (0, 0, wgetblockwidth (dstblock) - 1, wgetblockheight (dstblock) - 1);
    
    if (x + width > bx) display = (bx + 1) - x;
    else display = width;
    maxy = y + height - 1;
    if (maxy > by) maxy = by;
    if (y < ty) 
    {
        pr += (ty - y) << 4;   y = ty;  
    }                                  // clip y
    if (x < tx) 
    {
        pr += tx - x; display -= tx - x;   x = tx;  
    }                                  // clip x
    if (display > 0)
    {
	dst = dstblock + 4 + y*screenwidth[currentwindow] + x;
	for (ctr = y; ctr <= maxy; ctr++) 
	{
	    memcpy (dst, pr, display);
	    dst += screenwidth[currentwindow];
            pr += 16;
        }
    }
}

/***================================================================***
*             Put a block using either copy or xray modes             *
*	      but allows putting onto any size block, not just	      *
*	      full screen 320x200 blocks (used for side tiles	      *
*	      and main scrolling screens			      *
****================================================================***/
void wputanywhere (int x, int y, block pr, block sideptr, int method)
{
    int offset,width,height,display,maxy;
    int dwidth,dheight;
    int ctr;
    block dst;
    unsigned char t;
    
    dst = pr;
    t = *dst; dst++; width = t;
    t = *dst; dst++; width += t*256;   // get width and height of block
    t = *dst; dst++; height = t;       // of block to put anywhere
    t = *dst; dst++; height += t*256;
    
    dst = sideptr;
    t = *dst; dst++; dwidth = t;
    t = *dst; dst++; dwidth += t*256;  // get width and height of block
    t = *dst; dst++; dheight = t;      // that is being put onto (dest)
    t = *dst; dst++; dheight += t*256;
    
    pr += 4;
    wclip (0, 0, dwidth - 1, dheight - 1);
    
    
    if (x + width > bx) display = (bx + 1) - x;
    else display = width;
    maxy = y + height - 1;
    if (maxy > by) maxy = by;
    if (maxy > dheight) maxy = dheight;
    if (y < ty) 
    {
        pr += (ty - y) << 4;   y = ty;  
    }                                  // clip y
    if (x < tx) 
    {
        pr += tx - x; display -= tx - x;   x = tx;  
    }                                  // clip x
    if ((method == 0) & (display > 0))
    {
        dst = sideptr + 4 + y*dwidth + x;
        for (ctr = y; ctr <= maxy; ctr++) 
        {
            memcpy (dst, pr, display);
            dst += dwidth;
            pr += width;
        }
    }
    else if ((method == 1) & (display > 0))
    {
        dst = sideptr + 4 + y*dwidth + x;
        asm push ds;
        asm cld;
        for (ctr = y; ctr <= maxy; ctr++)
        {
            asm mov cx,display;
            asm lds si,pr;
            asm les di,dst;
            xloop:
            ;
            asm lodsb;
            asm cmp al,0;
            asm je ok;
            asm mov es:di,al;
            ok:
            ;
            asm inc di;
            asm loop xloop
                dst += dwidth;
            pr += width;
        }
        asm pop ds;
    }
}

